"""FitnessCalculatorConfiguration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from copy import deepcopy

from bluepyopt.ephys.locations import EPhysLocInstantiateException

from bluepyemodel.evaluation.efeature_configuration import EFeatureConfiguration
from bluepyemodel.evaluation.evaluator import LEGACY_PRE_PROTOCOLS
from bluepyemodel.evaluation.evaluator import PRE_PROTOCOLS
from bluepyemodel.evaluation.evaluator import define_location
from bluepyemodel.evaluation.evaluator import seclist_to_sec
from bluepyemodel.evaluation.protocol_configuration import ProtocolConfiguration
from bluepyemodel.tools.utils import are_same_protocol
from bluepyemodel.tools.utils import get_mapped_protocol_name

logger = logging.getLogger(__name__)


def _set_morphology_dependent_locations(recording, cell):
    """Here we deal with morphology dependent locations"""

    def _get_rec(recording, sec_id):
        new_rec = deepcopy(recording)
        rec_split = recording["name"].split(".")
        new_rec["type"] = "nrnseclistcomp"
        new_rec["name"] = f"{'.'.join(rec_split[:-1])}_{sec_id}.{rec_split[-1]}"
        new_rec["sec_index"] = sec_id
        return new_rec

    new_recs = []
    if recording["type"] == "somadistanceapic":
        new_recs = [deepcopy(recording)]
        new_recs[0]["sec_name"] = seclist_to_sec.get(
            recording["seclist_name"], recording["seclist_name"]
        )

    elif recording["type"] == "terminal_sections":
        # all terminal sections
        for sec_id, section in enumerate(getattr(cell.icell, recording["seclist_name"])):
            if len(section.subtree()) == 1:
                new_recs.append(_get_rec(recording, sec_id))

    elif recording["type"] == "all_sections":
        # all section of given type
        for sec_id, section in enumerate(getattr(cell.icell, recording["seclist_name"])):
            new_recs.append(_get_rec(recording, sec_id))

    else:
        new_recs = [deepcopy(recording)]

    if len(new_recs) == 0 and recording["type"] in [
        "somadistanceapic",
        "terminal_sections",
        "all_sections",
    ]:
        logger.warning("We could not add a location for %s", recording)
    return new_recs


class FitnessCalculatorConfiguration:
    """The goal of this class is to store the results of an efeature extraction (efeatures
    and protocols) or to contain the results of a previous extraction retrieved from an access
    point. This object is used for the creation of the fitness calculator.
    """

    def __init__(
        self,
        efeatures=None,
        protocols=None,
        name_rmp_protocol=None,
        name_rin_protocol=None,
        threshold_efeature_std=None,
        default_std_value=1e-3,
        validation_protocols=None,
        stochasticity=False,
        ion_variables=None,
        workflow_id=None,
    ):
        """Init.

        The arguments efeatures and protocols are expected to be in the format used for the
        storage of the fitness calculator configuration. To store the results of an extraction,
        use the method init_from_bluepyefe.

        Args:
            efeatures (list of dict): contains the description of the efeatures of the model
                in the format returned by the method as_dict of the EFeatureConfiguration class.
                Each unpacked dict will also act as the input to the EFeatureConfiguration class,
                details can be found in that class docstring. Below an example:

                .. code-block::

                    [
                        {
                            "efel_feature_name": "Spikecount",
                            "protocol_name": "IDrest_130",
                            "recording_name": "soma.v",
                            "mean": 6.026,
                            "original_std": 4.016,
                            "efeature_name": "Spikecount",
                            "weight": 1.0,  # optional
                            "efel_settings": {
                                "strict_stiminterval": true,
                                "Threshold": -30.0,
                                "interp_step": 0.025,
                                "stim_start": 700.0,
                                "stim_end": 2700.0
                            }
                        }
                    ]

            protocols (list of dict): contains the description of the protocols of the model
                in the format returned by the method as_dict of the ProtocolConfiguration class.
                Each unpacked dict will also act as the input to the ProtocolConfiguration class,
                details can be found in that class docstring. Below an example:

                .. code-block::

                    [
                        {
                            "name": "IDrest_130",
                            "stimuli": [
                                {
                                "delay": 663.3473451327434,
                                "amp": 0.3491827776582547,
                                "thresh_perc": 130.56103157894464,
                                "duration": 2053.050884955752,
                                "totduration": 3000.0,
                                "holding_current": -0.13307330411636328
                                }
                            ],
                            "recordings_from_config": [
                                {
                                "type": "CompRecording",
                                "name": "IDrest_130.soma.v",
                                "location": "soma",
                                "variable": "v"
                                }
                            ],
                            "validation": false,
                            "protocol_type": "ThresholdBasedProtocol",
                            "stochasticity": false
                        }
                    ]

            name_rmp_protocol (str or list): name and amplitude of protocol
                whose features are to be used as targets for the search of the RMP.
                e.g: ``["IV", 0]`` or ``"IV_0"``
            name_rin_protocol (str or list): name and amplitude of protocol
                whose features are to be used as targets for the search of the Rin.
                e.g: ``["IV", -20]`` or ``"IV_-20"``
            threshold_efeature_std (float): lower limit for the std expressed as a percentage of
                the mean of the features value (optional). Legacy.
             default_std_value (float): during and after extraction, this value will be used
                to replace the standard deviation if the standard deviation is 0.
            validation_protocols (list of str): name of the protocols used for validation only.
            stochasticity (bool or list of str): should channels behave stochastically if they can.
                If a list of protocol names is provided, the runs will be stochastic
                for these protocols, and deterministic for the other ones.
            ion_variables (list of str): ion current names and ionic concentration anmes
                for all available mechanisms
            workflow_id (str): EModelWorkflow id on nexus.
        """

        self.rmp_duration = 500.0
        self.rin_step_delay = 500.0
        self.rin_step_duration = 500.0
        self.rin_step_amp = -0.02
        self.rin_totduration = 1000.0
        self.search_holding_duration = 500.0
        self.search_threshold_step_delay = 500.0
        self.search_threshold_step_duration = 2000.0
        self.search_threshold_totduration = 3000.0
        self.ion_variables = ion_variables

        if protocols is None:
            self.protocols = []
        else:
            self.protocols = self.initialise_protocols(protocols)

        self.efeatures = []
        if efeatures is not None:
            self.efeatures = self.initialise_efeatures(
                efeatures, threshold_efeature_std, default_std_value
            )

        if validation_protocols is None:
            self.validation_protocols = []
        else:
            self.validation_protocols = validation_protocols

        self.stochasticity = stochasticity

        self.name_rmp_protocol = name_rmp_protocol
        self.name_rin_protocol = name_rin_protocol

        self.workflow_id = workflow_id
        self.default_std_value = default_std_value

    def initialise_protocols(self, protocols):
        """Initialise protocols from the FitnessCalculatorConfiguration format."""
        if protocols is None:
            return []
        return [ProtocolConfiguration(**p, ion_variables=self.ion_variables) for p in protocols]

    def initialise_efeatures(self, efeatures, threshold_efeature_std=None, default_std_value=1e-3):
        """Initialise efeatures from the FitnessCalculatorConfiguration format."""
        if efeatures is None:
            return []
        configured_efeatures = []
        for f in efeatures:
            f_dict = deepcopy(f)
            f_dict.pop("threshold_efeature_std", None)
            f_dict.pop("default_std_value", None)
            configured_efeatures.append(
                EFeatureConfiguration(
                    **f_dict,
                    threshold_efeature_std=f.get("threshold_efeature_std", threshold_efeature_std),
                    default_std_value=f.get("default_std_value", default_std_value),
                )
            )
        return configured_efeatures

    def protocol_exist(self, protocol_name):
        return bool(p for p in self.protocols if p.name == protocol_name)

    def check_stochasticity(self, protocol_name):
        """Check if stochasticity should be active for a given protocol"""
        if isinstance(self.stochasticity, list):
            return protocol_name in self.stochasticity
        return self.stochasticity

    def _add_bluepyefe_protocol(self, protocol_name, protocol):
        """"""

        # By default include somatic recording
        recordings = [
            {
                "type": "CompRecording",
                "name": f"{protocol_name}.soma.v",
                "location": "soma",
                "variable": "v",
            }
        ]

        stimulus = deepcopy(protocol["step"])
        stimulus["holding_current"] = protocol["holding"]["amp"]

        validation = any(are_same_protocol(protocol_name, p) for p in self.validation_protocols)
        stochasticity = self.check_stochasticity(protocol_name)

        protocol_type = "Protocol"
        if self.name_rmp_protocol and self.name_rin_protocol:
            protocol_type = "ThresholdBasedProtocol"

        tmp_protocol = ProtocolConfiguration(
            name=protocol_name,
            stimuli=[stimulus],
            recordings_from_config=recordings,
            validation=validation,
            ion_variables=self.ion_variables,
            stochasticity=stochasticity,
            protocol_type=protocol_type,
        )

        self.protocols.append(tmp_protocol)

    def _add_bluepyefe_efeature(self, feature, protocol_name, recording, threshold_efeature_std):
        """"""

        recording_name = "soma.v" if recording == "soma" else recording

        tmp_feature = EFeatureConfiguration(
            efel_feature_name=feature["feature"],
            protocol_name=protocol_name,
            recording_name=recording_name,
            mean=feature["val"][0],
            std=feature["val"][1],
            efeature_name=feature.get("efeature_name", None),
            efel_settings=feature.get("efel_settings", {}),
            threshold_efeature_std=threshold_efeature_std,
            sample_size=feature.get("n", None),
            default_std_value=self.default_std_value,
        )

        if (
            are_same_protocol(self.name_rmp_protocol, protocol_name)
            and feature["feature"] == "voltage_base"
        ):
            tmp_feature.protocol_name = "RMPProtocol"
            tmp_feature.efel_feature_name = "steady_state_voltage_stimend"
        if (
            are_same_protocol(self.name_rin_protocol, protocol_name)
            and feature["feature"] == "voltage_base"
        ):
            tmp_feature.protocol_name = "SearchHoldingCurrent"
            tmp_feature.efel_feature_name = "steady_state_voltage_stimend"
        if (
            are_same_protocol(self.name_rin_protocol, protocol_name)
            and feature["feature"] == "ohmic_input_resistance_vb_ssse"
        ):
            tmp_feature.protocol_name = "RinProtocol"

        if protocol_name not in PRE_PROTOCOLS and not self.protocol_exist(protocol_name):
            raise ValueError(
                f"Trying to register efeatures for protocol {protocol_name},"
                " but this protocol does not exist"
            )

        self.efeatures.append(tmp_feature)

    def init_from_bluepyefe(
        self,
        efeatures,
        protocols,
        currents,
        threshold_efeature_std,
        protocols_mapping=None,
    ):
        """Fill the configuration using the output of BluePyEfe"""

        if self.name_rmp_protocol and not any(
            are_same_protocol(self.name_rmp_protocol, p) for p in efeatures
        ):
            raise ValueError(
                f"The stimulus {self.name_rmp_protocol} requested for RMP "
                "computation couldn't be extracted from the ephys data."
            )
        if self.name_rin_protocol and not any(
            are_same_protocol(self.name_rin_protocol, p) for p in efeatures
        ):
            raise ValueError(
                f"The stimulus {self.name_rin_protocol} requested for Rin "
                "computation couldn't be extracted from the ephys data."
            )

        self.protocols = []
        self.efeatures = []

        self.validation_protocols = [
            get_mapped_protocol_name(vp, protocols_mapping) for vp in self.validation_protocols
        ]

        for protocol_name, protocol in protocols.items():
            p_name = get_mapped_protocol_name(protocol_name, protocols_mapping)
            self._add_bluepyefe_protocol(p_name, protocol)

        for protocol_name in efeatures:
            for recording in efeatures[protocol_name]:
                for feature in efeatures[protocol_name][recording]:
                    p_name = get_mapped_protocol_name(protocol_name, protocols_mapping)
                    self._add_bluepyefe_efeature(feature, p_name, recording, threshold_efeature_std)

        # Add the current related features
        if currents and self.name_rmp_protocol and self.name_rin_protocol:
            self.efeatures.append(
                EFeatureConfiguration(
                    efel_feature_name="bpo_holding_current",
                    protocol_name="SearchHoldingCurrent",
                    recording_name="soma.v",
                    mean=currents["holding_current"][0],
                    std=currents["holding_current"][1],
                    threshold_efeature_std=threshold_efeature_std,
                    default_std_value=self.default_std_value,
                )
            )

            self.efeatures.append(
                EFeatureConfiguration(
                    efel_feature_name="bpo_threshold_current",
                    protocol_name="SearchThresholdCurrent",
                    recording_name="soma.v",
                    mean=currents["threshold_current"][0],
                    std=currents["threshold_current"][1],
                    threshold_efeature_std=threshold_efeature_std,
                    default_std_value=self.default_std_value,
                )
            )

        self.remove_featureless_protocols()

    def _add_legacy_protocol(self, protocol_name, protocol):
        """"""

        # By default include somatic recording
        recordings = [
            {
                "type": "CompRecording",
                "name": f"{protocol_name}.soma.v",
                "location": "soma",
                "variable": "v",
            }
        ]

        if "extra_recordings" in protocol:
            for protocol_def in protocol["extra_recordings"]:
                recordings.append(protocol_def)
                protocol_def["name"] = (
                    f"{protocol_name}.{protocol_def['name']}.{protocol_def['var']}"
                )

        stimulus = deepcopy(protocol["stimuli"]["step"])
        if "holding" in protocol["stimuli"]:
            stimulus["holding_current"] = protocol["stimuli"]["holding"]["amp"]
        else:
            stimulus["holding_current"] = None

        validation = any(are_same_protocol(protocol_name, p) for p in self.validation_protocols)
        stochasticity = self.check_stochasticity(protocol_name)

        protocol_type = "Protocol"
        if "type" in protocol and protocol["type"] == "StepThresholdProtocol":
            protocol_type = "ThresholdBasedProtocol"

        tmp_protocol = ProtocolConfiguration(
            name=protocol_name,
            stimuli=[stimulus],
            recordings_from_config=recordings,
            validation=validation,
            ion_variables=self.ion_variables,
            protocol_type=protocol_type,
            stochasticity=stochasticity,
        )

        self.protocols.append(tmp_protocol)

    def _add_legacy_efeature(self, feature, protocol_name, recording, threshold_efeature_std):
        """"""

        recording_name = "soma.v" if recording == "soma" else recording

        tmp_feature = EFeatureConfiguration(
            efel_feature_name=feature["feature"],
            protocol_name=protocol_name,
            recording_name=recording_name,
            mean=feature["val"][0],
            std=feature["val"][1],
            efeature_name=feature.get("efeature_name", None),
            efel_settings=feature.get("efel_settings", {}),
            threshold_efeature_std=threshold_efeature_std,
            default_std_value=self.default_std_value,
        )

        if protocol_name == "Rin":
            if feature["feature"] == "ohmic_input_resistance_vb_ssse":
                tmp_feature.protocol_name = "RinProtocol"
            elif feature["feature"] == "voltage_base":
                tmp_feature.protocol_name = "SearchHoldingCurrent"
                tmp_feature.efel_feature_name = "steady_state_voltage_stimend"
            else:
                return

        if protocol_name == "RMP":
            if feature["feature"] == "voltage_base":
                tmp_feature.protocol_name = "RMPProtocol"
                tmp_feature.efel_feature_name = "steady_state_voltage_stimend"
            elif feature["feature"] == "Spikecount":
                tmp_feature.protocol_name = "RMPProtocol"
                tmp_feature.efel_feature_name = "Spikecount"
            else:
                return

        if protocol_name == "RinHoldCurrent":
            tmp_feature.protocol_name = "SearchHoldingCurrent"
        if protocol_name == "Threshold":
            tmp_feature.protocol_name = "SearchThresholdCurrent"

        if protocol_name not in PRE_PROTOCOLS and not self.protocol_exist(protocol_name):
            raise ValueError(
                f"Trying to register efeatures for protocol {protocol_name},"
                " but this protocol does not exist"
            )

        self.efeatures.append(tmp_feature)

    def init_from_legacy_dict(self, efeatures, protocols, threshold_efeature_std):
        self.protocols = []
        self.efeatures = []

        if (
            self.name_rmp_protocol
            and not any(are_same_protocol(self.name_rmp_protocol, p) for p in efeatures)
            and "RMP" not in efeatures
        ):
            raise ValueError(
                f"The protocol {self.name_rmp_protocol} requested for RMP nor RMPProtocol "
                "are present in your efeatures json file."
            )

        if (
            self.name_rin_protocol
            and not any(are_same_protocol(self.name_rin_protocol, p) for p in efeatures)
            and "Rin" not in efeatures
        ):
            raise ValueError(
                f"The protocol {self.name_rin_protocol} requested for Rin nor RinProtocol "
                "are present in your efeatures json file."
            )

        for protocol_name, protocol in protocols.items():
            if protocol_name == "RMP":
                self.rmp_duration = protocol["stimuli"]["step"]["duration"]
            if protocol_name == "Rin":
                self.rin_step_delay = protocol["stimuli"]["step"]["delay"]
                self.rin_step_duration = protocol["stimuli"]["step"]["duration"]
                self.rin_step_amp = protocol["stimuli"]["step"]["amp"]
                self.rin_totduration = protocol["stimuli"]["step"]["totduration"]
            if protocol_name == "ThresholdDetection":
                self.search_threshold_step_delay = protocol["step_template"]["stimuli"]["step"][
                    "delay"
                ]
                self.search_threshold_step_duration = protocol["step_template"]["stimuli"]["step"][
                    "duration"
                ]
                self.search_threshold_totduration = protocol["step_template"]["stimuli"]["step"][
                    "totduration"
                ]

            if protocol_name in PRE_PROTOCOLS + LEGACY_PRE_PROTOCOLS:
                continue

            self._add_legacy_protocol(protocol_name, protocol)
            validation = protocol.get("validation", False)

            if validation != self.protocols[-1].validation:
                raise ValueError(
                    "The protocol was set as a validation protocol in the json but is not present "
                    "as a validation protocol in the settings"
                )

        for protocol_name in efeatures:
            for recording in efeatures[protocol_name]:
                for feature in efeatures[protocol_name][recording]:
                    self._add_legacy_efeature(
                        feature, protocol_name, recording, threshold_efeature_std
                    )

        self.remove_featureless_protocols()

    def remove_featureless_protocols(self):
        """Remove the protocols that do not have any matching efeatures"""

        to_remove = []

        for i, protocol in enumerate(self.protocols):
            for efeature in self.efeatures:
                if efeature.protocol_name == protocol.name:
                    break
            else:
                to_remove.append(i)

        self.protocols = [p for i, p in enumerate(self.protocols) if i not in to_remove]

    def configure_morphology_dependent_locations(self, _cell, simulator):
        """"""

        cell = deepcopy(_cell)
        cell.params = None
        cell.mechanisms = None
        cell.instantiate(sim=simulator)

        # TODO: THE SAME FOR STIMULI

        skipped_recordings = []
        for i, protocol in enumerate(self.protocols):
            recordings = []
            for j, rec in enumerate(protocol.recordings):
                if rec["type"] != "CompRecording":
                    for _rec in _set_morphology_dependent_locations(rec, cell):
                        try:
                            location = define_location(_rec)
                            location.instantiate(sim=simulator, icell=cell.icell)
                            recordings.append(_rec)
                        except EPhysLocInstantiateException:
                            logger.warning(
                                "Could not find %s, ignoring recording at this location",
                                location.name,
                            )
                            skipped_recordings.append(_rec["name"])
                else:
                    recordings.append(self.protocols[i].recordings[j])
            self.protocols[i].recordings = recordings

        to_remove = []
        efeatures = []
        for i, efeature in enumerate(self.efeatures):
            if isinstance(efeature.recording_name, str):
                # remove efeature associated to skipped recording
                for skiprec in skipped_recordings:
                    if f"{efeature.protocol_name}.{efeature.recording_name}" == skiprec:
                        to_remove.append(i)
                        logger.warning("Removing %s", efeature.name)
                        continue
                # if the loc of the recording is of the form axon*.v, we replace * by
                # all the corresponding int from the created recordings
                loc_name, rec_name = efeature.recording_name.split(".")
                if loc_name[-1] == "*":
                    to_remove.append(i)
                    protocol = next(p for p in self.protocols if p.name == efeature.protocol_name)
                    for rec in protocol.recordings:
                        base_rec_name = rec["name"].split(".")[1]
                        if base_rec_name.startswith(loc_name[:-1]):
                            efeatures.append(deepcopy(efeature))
                            efeatures[-1].recording_name = f"{base_rec_name}.{rec_name}"

        self.efeatures = [f for i, f in enumerate(self.efeatures) if i not in to_remove] + efeatures

    def get_related_nexus_ids(self):
        return {
            "generation": {
                "type": "Generation",
                "activity": {
                    "type": "Activity",
                    "followedWorkflow": {
                        "type": "EModelWorkflow",
                        "id": self.workflow_id,
                    },
                },
            }
        }

    def as_dict(self):
        """Used for the storage of the configuration"""

        return {
            "efeatures": [e.as_dict() for e in self.efeatures],
            "protocols": [p.as_dict() for p in self.protocols],
        }

    def __str__(self):
        """String representation"""

        str_form = "Fitness Calculator Configuration:\n\n"

        str_form += "Protocols:\n"
        for p in self.protocols:
            str_form += f"   {p.as_dict()}\n"

        str_form += "EFeatures:\n"
        for f in self.efeatures:
            str_form += f"   {f.as_dict()}\n"

        return str_form



"""LocalAccessPoint class."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import glob
import json
import logging
from functools import cached_property
from itertools import chain
from pathlib import Path

import fasteners
from bluepyefe.tools import NumpyEncoder

from bluepyemodel.access_point.access_point import DataAccessPoint
from bluepyemodel.efeatures_extraction.targets_configuration import TargetsConfiguration
from bluepyemodel.emodel_pipeline.emodel import EModel
from bluepyemodel.emodel_pipeline.emodel_metadata import EModelMetadata
from bluepyemodel.emodel_pipeline.emodel_settings import EModelPipelineSettings
from bluepyemodel.emodel_pipeline.emodel_workflow import EModelWorkflow
from bluepyemodel.evaluation.evaluator import LEGACY_PRE_PROTOCOLS
from bluepyemodel.evaluation.evaluator import PRE_PROTOCOLS
from bluepyemodel.evaluation.fitness_calculator_configuration import FitnessCalculatorConfiguration
from bluepyemodel.export_emodel.utils import get_hoc_file_path
from bluepyemodel.export_emodel.utils import get_output_path_from_metadata
from bluepyemodel.model.mechanism_configuration import MechanismConfiguration
from bluepyemodel.model.neuron_model_configuration import NeuronModelConfiguration
from bluepyemodel.tools.mechanisms import get_mechanism_currents
from bluepyemodel.tools.mechanisms import get_mechanism_name

logger = logging.getLogger(__name__)

seclist_to_sec = {
    "somatic": "soma",
    "apical": "apic",
    "axonal": "axon",
    "myelinated": "myelin",
}

SUPPORTED_MORPHOLOGY_EXTENSIONS = (".asc", ".swc", ".ASC", ".SWC")


class LocalAccessPoint(DataAccessPoint):
    """Access point to access configuration files and e-models when stored locally."""

    def __init__(
        self,
        emodel,
        emodel_dir=None,
        etype=None,
        ttype=None,
        mtype=None,
        species=None,
        brain_region=None,
        iteration_tag=None,
        synapse_class=None,
        final_path=None,
        recipes_path=None,
        legacy_dir_structure=False,
        with_seeds=False,
    ):
        """Init

        Args:
            emodel (str): name of the emodel.
            emodel_dir (str): path to the working directory.
                Default to current working directory.
                If iteration_tag is not None, it will be ./run/iteration_tag
            ttype (str): name of the t-type
            iteration_tag (str): iteration tag
            final_path (str): path to final.json which will contain the optimised models.
            recipes_path (str): path to the json file which should contain the path to the
                configuration files used for each etype. The content of this file should follow the
                format:

                .. code-block::

                    {
                        "emodelname": {
                            "morph_path": "./morphologies/",
                            "morphology": [[morphologyname", "morphfile.asc"]],
                            "params": "config/params/pyr.json",
                            "protocol": "config/protocols/emodelname.json",
                            "features": "config/features/emodelname.json"
                        }
                    }
            legacy_dir_structure (bool): if true, the path coming from the recipes will be replaced
                by "self.emodel_dir / emodel / json_path_from_recipes". To be deprecated.
            with_seed (bool): allows for emodel_seed type of emodel names in final.json
        """

        # pylint: disable=too-many-arguments

        super().__init__(
            emodel,
            etype,
            ttype,
            mtype,
            species,
            brain_region,
            iteration_tag,
            synapse_class,
        )

        if emodel_dir is None:
            self.emodel_dir = Path.cwd()
            if iteration_tag:
                self.emodel_dir = self.emodel_dir / "run" / iteration_tag
        else:
            self.emodel_dir = Path(emodel_dir)

        self.recipes_path = recipes_path
        self.legacy_dir_structure = legacy_dir_structure
        self.with_seeds = with_seeds

        if final_path is None:
            self.final_path = self.emodel_dir / "final.json"
        else:
            self.final_path = Path(final_path)

        # variable to set if one wants to save a new final path after optimisation
        self.new_final_path = None

        Path(".tmp/").mkdir(exist_ok=True)
        self.rw_lock_final = fasteners.InterProcessReaderWriterLock(".tmp/final.lock")
        self.rw_lock_final_tmp = fasteners.InterProcessReaderWriterLock(".tmp/final_tmp.lock")

        self.pipeline_settings = self.load_pipeline_settings()
        self.unfrozen_params = None

    @cached_property
    def morph_dir(self):
        """Return the morphology directory as read from the recipes, or fallback to 'morphology'."""
        recipes = self.get_recipes()
        return Path(self.emodel_dir, recipes.get("morph_path", "morphology"))

    @cached_property
    def morph_path(self):
        """Return the path to the morphology file as read from the recipes."""
        recipes = self.get_recipes()

        morph_file = None
        if isinstance(recipes["morphology"], str):
            morph_file = recipes["morphology"]
        else:
            for _, morph_file in recipes["morphology"]:
                if morph_file.endswith(SUPPORTED_MORPHOLOGY_EXTENSIONS):
                    break

        if not morph_file or not morph_file.endswith(SUPPORTED_MORPHOLOGY_EXTENSIONS):
            raise FileNotFoundError(f"Morphology file not defined or not supported: {morph_file}")

        morph_path = self.morph_dir / morph_file

        if not morph_path.is_file():
            raise FileNotFoundError(f"Morphology file not found: {morph_path}")
        if str(Path.cwd()) not in str(morph_path.resolve()) and self.emodel_metadata.iteration:
            raise FileNotFoundError(
                "When using a githash or iteration tag, the path to the morphology must be local"
                " otherwise it cannot be archived during the creation of the githash. To solve"
                " this issue, you can copy the morphology from "
                f"{morph_path.resolve()} to {Path.cwd() / 'morphologies'} and update your "
                "recipes."
            )

        return morph_path

    def set_emodel(self, emodel):
        """Setter for the name of the emodel, check it exists (with or without seed) in recipe."""
        _emodel = "_".join(emodel.split("_")[:2]) if self.with_seeds else emodel
        if _emodel not in self.get_all_recipes():
            raise ValueError(
                f"Cannot set the emodel name to {_emodel} which does not exist in the recipes."
            )

        self.unfrozen_params = None
        self.emodel_metadata.emodel = _emodel

    def load_pipeline_settings(self):
        """ """
        recipes = self.get_recipes()
        settings = recipes.get("pipeline_settings", {})
        if isinstance(settings, str):
            # read the pipeline settings from file
            settings = self.get_json("pipeline_settings")
        if "morph_modifiers" not in settings:
            settings["morph_modifiers"] = recipes.get("morph_modifiers", None)
        return EModelPipelineSettings(**settings)

    def _config_to_final(self, config):
        """Convert the configuration stored in EM_*.json to the format used for final.json."""
        return {
            self.emodel_metadata.emodel: {
                **vars(self.emodel_metadata),
                "score": config["fitness"],  # float
                "parameters": config["parameter"],  # list[dict]
                "fitness": config["score"],  # list[dict]
                "features": config["features"],  # list[dict]
                "validation_fitness": config["scoreValidation"],  # list[dict]
                "validated": config["passedValidation"],  # bool
                "seed": config["seed"],  # int
            }
        }

    def get_final_content(self, lock_file=True):
        """Return the final content from recipes if available, or fallback to final.json"""
        recipes = self.get_recipes()
        if "final" in recipes:
            if self.final_path and self.final_path.is_file():
                logger.warning("Ignored %s, using file from recipes", self.final_path)
            data = self.get_json("final")
            return self._config_to_final(data)
        return self.get_final(lock_file=lock_file)

    def get_final(self, lock_file=True):
        """Get emodel dictionary from final.json."""
        if self.final_path is None:
            raise TypeError("Final_path is None")

        if not self.final_path.is_file():
            logger.info("%s does not exist and will be created", self.final_path)
            return {}

        try:
            if lock_file:
                self.rw_lock_final.acquire_read_lock()

            with open(self.final_path, "r") as f:
                final = json.load(f)

            if lock_file:
                self.rw_lock_final.release_read_lock()
        except json.decoder.JSONDecodeError:
            try:
                if lock_file:
                    self.rw_lock_final_tmp.acquire_read_lock()
                _tmp_final_path = self.final_path.with_name(
                    self.final_path.stem + "_tmp" + self.final_path.suffix
                )
                with open(_tmp_final_path, "r") as f:
                    final = json.load(f)

                if lock_file:
                    self.rw_lock_final_tmp.release_read_lock()
            except (json.decoder.JSONDecodeError, FileNotFoundError):
                logger.error("Cannot load final. final.json does not exist or is corrupted.")

        return final

    def save_final(self, final, final_path, lock_file=True):
        """Save final emodels in json"""

        if final_path is None:
            raise TypeError("Final_path is None")

        if lock_file:
            self.rw_lock_final.acquire_write_lock()

        with open(final_path, "w+") as fp:
            json.dump(final, fp, indent=2)

        if lock_file:
            self.rw_lock_final.release_write_lock()
            self.rw_lock_final_tmp.acquire_write_lock()

        with open(final_path.with_name(final_path.stem + "_tmp.json"), "w+") as fp:
            json.dump(final, fp, indent=2)

        if lock_file:
            self.rw_lock_final_tmp.release_write_lock()

    def get_all_recipes(self):
        """Load the recipes from a json file.

        See docstring of __init__ for the format of the file of recipes.
        """

        if self.legacy_dir_structure:
            emodel = "_".join(self.emodel_metadata.emodel.split("_")[:2])
            recipes_path = self.emodel_dir / emodel / "config" / "recipes" / "recipes.json"
        else:
            recipes_path = self.emodel_dir / self.recipes_path

        with open(recipes_path, "r") as f:
            return json.load(f)

    def get_recipes(self):
        """Load the recipes from a json file for an emodel.

        See docstring of __init__ for the format of the file of recipes.
        """
        if self.with_seeds:
            emodel = "_".join(self.emodel_metadata.emodel.split("_")[:2])
        else:
            emodel = self.emodel_metadata.emodel
        return self.get_all_recipes()[emodel]

    def get_json_path(self, recipe_entry):
        """Helper function to get a json path in recipe."""

        json_path = Path(self.get_recipes()[recipe_entry])

        if self.legacy_dir_structure:
            emodel = "_".join(self.emodel_metadata.emodel.split("_")[:2])
            json_path = self.emodel_dir / emodel / json_path
        elif not json_path.is_absolute():
            json_path = self.emodel_dir / json_path
        return json_path

    def get_json(self, recipe_entry):
        """Helper function to load a json using path in recipe."""
        json_path = self.get_json_path(recipe_entry)
        with open(json_path, "r") as f:
            return json.load(f)

    def get_model_name_for_final(self, seed):
        """Return model name used as key in final.json."""

        if self.emodel_metadata.iteration:
            return f"{self.emodel_metadata.emodel}__{self.emodel_metadata.iteration}__{seed}"

        logger.warning(
            "The iteration is %s. It is strongly advised to use an iteration tag in the future.",
            self.emodel_metadata.iteration,
        )

        return f"{self.emodel_metadata.emodel}__{seed}"

    def store_emodel(self, emodel):
        """Store an emodel obtained from BluePyOpt in the final.json. Note that if a model in the
        final.json has the same key (emodel__iteration_tag__seed), it will be overwritten.
        """

        with self.rw_lock_final.write_lock():
            with self.rw_lock_final_tmp.write_lock():
                final = self.get_final(lock_file=False)
                model_name = self.get_model_name_for_final(emodel.seed)

                if model_name in final:
                    logger.warning(
                        "Entry %s was already in the final.json and will be overwritten",
                        model_name,
                    )

                pdf_dependencies = emodel.build_pdf_dependencies(int(emodel.seed))

                if self.new_final_path is None:
                    final_path = self.final_path
                    _params = emodel.parameters
                else:
                    if "params" in final[self.emodel_metadata.emodel]:
                        _params = final[self.emodel_metadata.emodel]["params"]
                    else:
                        _params = final[self.emodel_metadata.emodel]["parameters"]
                    _params.update(emodel.parameters)
                    final_path = self.new_final_path

                final[model_name] = vars(self.emodel_metadata)
                final[model_name].update(
                    {
                        "score": sum(list(emodel.scores.values())),
                        "parameters": _params,
                        "fitness": emodel.scores,
                        "features": emodel.features,
                        "validation_fitness": emodel.scores_validation,
                        "validated": emodel.passed_validation,
                        "seed": int(emodel.seed),
                        "pdfs": pdf_dependencies,
                    }
                )

                self.save_final(final, Path(final_path), lock_file=False)

    def set_unfrozen_params(self, params):
        """Freeze parameters for partial optimisation."""
        # todo: check if params exists in final.json
        self.unfrozen_params = params

    def _freeze_params(self, params):
        """Freeze parameters to final except these in self.unfrozen_params."""
        emodel_params = self.get_emodel()["parameters"]
        for loc in params:
            for param in params[loc]:
                name = ".".join([param["name"], loc])
                if name not in self.unfrozen_params and isinstance(param["val"], list):
                    param["val"] = emodel_params[name]

    def store_model_configuration(self, configuration, path=None):
        """Store a model configuration as a json"""

        if path is None:
            path = Path(self.get_recipes()["params"])

        path.parent.mkdir(parents=True, exist_ok=True)
        with open(path, "w") as f:
            json.dump(configuration.as_dict(), f, indent=2)

    def get_mechanisms_directory(self):
        """Return the path to the directory containing the mechanisms for the current emodel"""

        mechanisms_directory = self.emodel_dir / "mechanisms"

        if mechanisms_directory.is_dir():
            return mechanisms_directory.resolve()

        return None

    def get_available_mechanisms(self):
        """Get the list of names of the available mechanisms"""

        mech_dir = self.get_mechanisms_directory()
        if mech_dir is None:
            return None

        available_mechanisms = []
        for mech_file in glob.glob(str(Path(mech_dir) / "*.mod")):
            ion_currents, nonspecific_currents, ion_conc = get_mechanism_currents(mech_file)
            name = get_mechanism_name(mech_file)
            available_mechanisms.append(
                MechanismConfiguration(
                    name=name,
                    location=None,
                    ion_currents=ion_currents,
                    nonspecific_currents=nonspecific_currents,
                    ionic_concentrations=ion_conc,
                )
            )

        return available_mechanisms

    def get_available_morphologies(self):
        """Get the list of names of available morphologies"""
        morph_dir = self.morph_dir

        if not morph_dir.is_dir():
            return None

        patterns = ["*" + ext for ext in SUPPORTED_MORPHOLOGY_EXTENSIONS]
        return {morph_file.stem for pattern in patterns for morph_file in morph_dir.glob(pattern)}

    def get_model_configuration(self, skip_get_available_morph=False):
        """Get the configuration of the model, including parameters, mechanisms and distributions

        Args:
            skip_get_available_morph (bool): only used in nexus access point.
        """

        configuration = NeuronModelConfiguration(
            available_mechanisms=self.get_available_mechanisms(),
            available_morphologies=self.get_available_morphologies(),
        )

        try:
            parameters = self.get_json("parameters")
        except KeyError:
            parameters = self.get_json("params")

        if isinstance(parameters["parameters"], dict):
            parameters["parameters"].pop("__comment", None)

        if self.unfrozen_params is not None:
            self._freeze_params(parameters["parameters"])

        if isinstance(parameters["mechanisms"], dict):
            configuration.init_from_legacy_dict(parameters, self.get_morphologies())
        else:
            configuration.init_from_dict(parameters, self.get_morphologies())

        configuration.mapping_multilocation = self.get_recipes().get("multiloc_map", None)
        configuration.morph_modifiers = self.get_recipes().get("morph_modifiers", None)
        return configuration

    def store_targets_configuration(self, configuration):
        """Store the configuration of the targets (targets and ephys files used)"""

        config_dict = {
            "emodel": self.emodel_metadata.emodel,
            "ttype": self.emodel_metadata.ttype,
        }

        config_dict.update(configuration.as_dict())

        path_extract_config = self.pipeline_settings.path_extract_config
        Path(path_extract_config).parent.mkdir(parents=True, exist_ok=True)

        with open(str(path_extract_config), "w") as f:
            f.write(json.dumps(config_dict, indent=2, cls=NumpyEncoder))

    def get_targets_configuration(self):
        """Get the configuration of the targets (targets and ephys files used)"""

        path_extract_config = self.emodel_dir / self.pipeline_settings.path_extract_config

        with open(path_extract_config, "r") as f:
            config_dict = json.load(f)

        configuration = TargetsConfiguration(
            files=config_dict["files"],
            targets=config_dict["targets"],
            protocols_rheobase=config_dict["protocols_rheobase"],
            additional_fitness_efeatures=config_dict.get("additional_fitness_efeatures", None),
            additional_fitness_protocols=config_dict.get("additional_fitness_protocols", None),
            protocols_mapping=config_dict.get("protocols_mapping", None),
        )

        return configuration

    def store_fitness_calculator_configuration(self, configuration):
        """Store a fitness calculator configuration"""

        config_path = self.emodel_dir / self.get_recipes()["features"]
        config_path.parent.mkdir(parents=True, exist_ok=True)

        with open(str(config_path), "w") as f:
            f.write(json.dumps(configuration.as_dict(), indent=2, cls=NumpyEncoder))

    def get_fitness_calculator_configuration(self, record_ions_and_currents=False):
        """Get the configuration of the fitness calculator (efeatures and protocols)"""

        config_dict = self.get_json("features")

        legacy = "efeatures" not in config_dict and "protocols" not in config_dict

        # contains ion currents and ionic concentrations to be recorded
        ion_variables = None
        if record_ions_and_currents:
            ion_currents, ionic_concentrations = self.get_ion_currents_concentrations()
            if ion_currents is not None and ionic_concentrations is not None:
                ion_variables = list(chain.from_iterable((ion_currents, ionic_concentrations)))

        if legacy:
            efeatures = self.get_json("features")
            protocols = self.get_json("protocol")

            from_bpe = False
            for protocol_name, protocol in protocols.items():
                if protocol_name in PRE_PROTOCOLS + LEGACY_PRE_PROTOCOLS:
                    continue
                if "stimuli" not in protocol:
                    from_bpe = True
                    break

            configuration = FitnessCalculatorConfiguration(
                name_rmp_protocol=self.pipeline_settings.name_rmp_protocol,
                name_rin_protocol=self.pipeline_settings.name_Rin_protocol,
                validation_protocols=self.pipeline_settings.validation_protocols,
                stochasticity=self.pipeline_settings.stochasticity,
                ion_variables=ion_variables,
            )

            if from_bpe:
                configuration.init_from_bluepyefe(
                    efeatures,
                    protocols,
                    {},
                    self.pipeline_settings.threshold_efeature_std,
                )
            else:
                configuration.init_from_legacy_dict(
                    efeatures, protocols, self.pipeline_settings.threshold_efeature_std
                )

        else:
            configuration = FitnessCalculatorConfiguration(
                efeatures=config_dict["efeatures"],
                protocols=config_dict["protocols"],
                name_rmp_protocol=self.pipeline_settings.name_rmp_protocol,
                name_rin_protocol=self.pipeline_settings.name_Rin_protocol,
                validation_protocols=self.pipeline_settings.validation_protocols,
                stochasticity=self.pipeline_settings.stochasticity,
                ion_variables=ion_variables,
            )

        return configuration

    def create_emodel_workflow(self, state="not launched"):
        """Create an empty EModelWorkflow instance. EModel workflow should not be used in local"""
        return EModelWorkflow(
            None,
            None,
            None,
            state=state,
        )

    def get_emodel_workflow(self):
        """Emodel workflow is not used in local, so return None here"""
        return None, None

    def check_emodel_workflow_configurations(self, emodel_workflow):
        """Emodel workflow is not used in local, so always return True to let the workflow run"""
        # pylint: disable=unused-argument
        return True

    def store_or_update_emodel_workflow(self, emodel_workflow):
        """Emodel workflow is not used in local, so pass"""

    def get_morphologies(self):
        """Get the name and path to the morphologies from the recipes.

        Returns:
            morphology_definition (list): [{'name': morph_name, 'path': 'morph_path'}]. Might
            contain the additional entries "seclist_names" and "secarray_names" if they are
            present in the recipes.
        """

        recipes = self.get_recipes()
        morphology_definition = {
            "name": self.morph_path.stem,
            "path": str(self.morph_path),
        }
        if "seclist_names" in recipes:
            morphology_definition["seclist_names"] = recipes["seclist_names"]

        if "secarray_names" in recipes:
            morphology_definition["secarray_names"] = recipes["secarray_names"]

        return morphology_definition

    def format_emodel_data(self, model_data):
        """Format emodel data."""

        if "githash" in model_data:
            iteration_tag = model_data["githash"]
        elif "iteration_tag" in model_data:
            iteration_tag = model_data["iteration_tag"]
        elif "iteration" in model_data:
            iteration_tag = model_data["iteration"]
        else:
            iteration_tag = None

        emodel_metadata = EModelMetadata(
            emodel=model_data.get("emodel", self.emodel_metadata.emodel),
            etype=model_data.get("etype", None),
            ttype=model_data.get("ttype", None),
            mtype=model_data.get("mtype", None),
            species=model_data.get("species", None),
            brain_region=model_data.get("brain_region", None),
            iteration_tag=iteration_tag,
            synapse_class=model_data.get("synapse_class", None),
        )

        emodel = EModel(
            fitness=model_data.get("score", None),
            parameter=model_data.get("params", model_data.get("parameters", None)),
            score=model_data.get("fitness", None),
            features=model_data.get("features", None),
            scoreValidation=model_data.get("validation_fitness", None),
            passedValidation=model_data.get("validated", None),
            seed=model_data.get("seed", None),
            emodel_metadata=emodel_metadata,
        )

        return emodel

    def get_emodel(self, lock_file=True):
        """Get dict with parameter of single emodel (including seed if any)"""

        final = self.get_final_content(lock_file=lock_file)

        if self.emodel_metadata.emodel in final:
            return self.format_emodel_data(final[self.emodel_metadata.emodel])

        logger.warning("Could not find models for emodel %s", self.emodel_metadata.emodel)

        return None

    def get_emodels(self, emodels=None):
        """Get a list of emodels

        Args:
            emodels (list): list of names of the emodels.
        """

        if emodels is None:
            emodels = [self.emodel_metadata.emodel]

        models = []
        for mod_data in self.get_final_content().values():
            if mod_data["emodel"] in emodels:
                models.append(self.format_emodel_data(mod_data))

        filtered_models = []
        api_metadata = self.emodel_metadata.as_dict()
        for m in models:
            model_metadata = m.emodel_metadata.as_dict()
            for f, v in api_metadata.items():
                if f in model_metadata and v != model_metadata[f] and f != "allen_notation":
                    break
            else:
                filtered_models.append(m)

        return filtered_models

    def has_pipeline_settings(self):
        """Returns True if pipeline settings are present in the recipes"""

        return "pipeline_settings" in self.get_recipes()

    def has_fitness_calculator_configuration(self):
        """Check if the fitness calculator configuration exists"""

        recipes = self.get_recipes()

        return Path(recipes["features"]).is_file()

    def has_targets_configuration(self):
        """Check if the target configuration exists"""

        return (
            self.pipeline_settings.path_extract_config
            and Path(self.pipeline_settings.path_extract_config).is_file()
        )

    def has_model_configuration(self):
        """Check if the model configuration exists"""

        recipes = self.get_recipes()

        return Path(recipes["params"]).is_file()

    def get_emodel_etype_map(self):
        final = self.get_final_content()
        return {emodel: emodel.split("_")[0] for emodel in final}

    def get_emodel_names(self):
        """Get the list of all the names of emodels

        Returns:
            dict: keys are emodel names with seed, values are names without seed.
        """

        final = self.get_final_content()

        return {mod_name: mod.get("emodel", mod_name) for mod_name, mod in final.items()}

    def has_best_model(self, seed):
        """Check if the best model has been stored."""

        final = self.get_final_content()

        model_name = self.get_model_name_for_final(seed)

        return model_name in final

    def is_checked_by_validation(self, seed):
        """Check if the emodel with a given seed has been checked by Validation task."""

        final = self.get_final_content()

        model_name = self.get_model_name_for_final(seed)

        model = final.get(model_name, {})
        if "validated" in model and model["validated"] is not None:
            return True

        return False

    def is_validated(self):
        """Check if enough models have been validated."""

        n_validated = 0
        final = self.get_final_content()

        for _, entry in final.items():
            if (
                entry["iteration"] == self.emodel_metadata.iteration
                and entry["emodel"] == self.emodel_metadata.emodel
                and entry["validated"]
            ):
                n_validated += 1

        return n_validated >= self.pipeline_settings.n_model

    @classmethod
    def add_entry_recipes(
        cls,
        recipes_path,
        emodel,
        morph_path,
        morphology,
        parameters_path,
        protocols_path,
        features_path,
        pipeline_settings=None,
    ):
        """Append an entry to the recipes, create the file if it does not exist

        Args:
            recipes_path (str): path to the json containig the recipes.
            emodel (str): name of the emodel.
            morph_path (str): path the directory containing the morphologies.
            morphology (list of str): name of the morphology and name of the file.
            parameters_path (str): path to the json that contains the parameters.
            protocols_path (str): path to the json that contains the protocols.
            features_path (str): path to the json that contains the features.
            pipeline_settings (dict): pipeline settings.
        """

        recipes_path = Path(recipes_path)
        recipes_path.parent.mkdir(parents=True, exist_ok=True)

        if recipes_path.is_file():
            with recipes_path.open("r") as f:
                recipes = json.load(f)
        else:
            recipes = {}

        if pipeline_settings is None:
            settings = EModelPipelineSettings().as_dict()
        else:
            settings = pipeline_settings

        recipes[emodel] = {
            "morph_path": morph_path,
            "morphology": [morphology],
            "params": parameters_path,
            "protocol": protocols_path,
            "features": features_path,
            "pipeline_settings": settings,
        }

        with recipes_path.open("w") as f:
            json.dump(recipes, f, indent=2)

    def store_hocs(
        self,
        only_validated=False,
        only_best=True,
        seeds=None,
        map_function=map,
        new_emodel_name=None,
        description=None,
        output_base_dir="export_emodels_hoc",
    ):
        """Not Implemented"""
        raise NotImplementedError

    def store_emodels_hoc(
        self,
        only_validated=False,
        only_best=True,
        seeds=None,
        map_function=map,
        new_emodel_name=None,
        description=None,
    ):
        """Not Implemented"""
        raise NotImplementedError

    def store_emodels_sonata(
        self,
        only_validated=False,
        only_best=True,
        seeds=None,
        map_function=map,
        new_emodel_name=None,
        description=None,
    ):
        """Not Implemented"""
        raise NotImplementedError

    def sonata_exists(self, seed):
        """Returns True if the sonata hoc file has been exported"""
        output_path = get_output_path_from_metadata(
            "export_emodels_sonata", self.emodel_metadata, seed
        )
        hoc_file_path = get_hoc_file_path(output_path)
        return Path(hoc_file_path).is_file()



"""Neuron Model Configuration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from bluepyemodel.model.distribution_configuration import DistributionConfiguration
from bluepyemodel.model.mechanism_configuration import MechanismConfiguration
from bluepyemodel.model.morphology_configuration import MorphologyConfiguration
from bluepyemodel.model.parameter_configuration import ParameterConfiguration
from bluepyemodel.model.utils import temp_ljp_check
from bluepyemodel.tools.mechanisms import NEURON_BUILTIN_MECHANISMS

logger = logging.getLogger(__name__)

multiloc_map = {
    "all": ["apical", "basal", "somatic", "axonal"],
    "alldend": ["apical", "basal"],
    "somadend": ["apical", "basal", "somatic"],
    "allnoaxon": ["apical", "basal", "somatic"],
    "somaxon": ["axonal", "somatic"],
    "allact": ["apical", "basal", "somatic", "axonal"],
}

global_parameters = ["v_init", "celsius", "cm", "Ra", "ena", "ek"]


class NeuronModelConfiguration:
    """A neuron model configuration, which includes the model's parameters, distributions,
    mechanisms and a morphology."""

    def __init__(
        self,
        parameters=None,
        mechanisms=None,
        distributions=None,
        morphology=None,
        available_mechanisms=None,
        available_morphologies=None,
        morph_modifiers=None,
        extra_mech_ids=None,
    ):
        """Creates a model configuration, which includes the model parameters, distributions,
        mechanisms and a morphology.

        WARNING: If you create a new NeuronModelConfiguration, do not specify the parameters,
        mechanisms, distributions and morphology here at instantiation. Instead, create an empty
        configuration and then use the class method: add_distribution, add_parameter, etc.
        Example::

            .. code-block::

                config = NeuronModelConfiguration()
                config.add_parameter(parameter_name, locations, value, mechanism)
                config.morphology = MorphologyConfiguration(morph_name, format=".swc")

        Args:
            parameters (list of dict): contains the description of the parameters of the model
                in the format returned by the method as_dict of the ParameterConfiguration class.
            mechanisms (list of dict): contains the description of the mechanisms of the model
                in the format returned by the method as_dict of the MechanismConfiguration class.
            distributions (list of dict): contains the description of the distributions of the model
                in the format returned by the method as_dict of the DistributionConfiguration class.
            morphology (dict):  contains the description of the morphology of the model in the
                format returned by the method as_dict of the morphology class.
            available_mechanisms (list of MechanismConfiguration): list of the names (
                and optionally versions) of the available mechanisms in the ``./mechanisms``
                directory for the local access point or on Nexus for the Nexus access point.
            available_morphologies (list of str): list of the names of the available morphology in
                the ``./morphology`` directory for the local access point or on Nexus for the
                Nexus access point.
            morph_modifiers (list of str or list of list):
                If str, name of the morph modifier to use from bluepyemodel.evaluation.modifiers.
                If List of morphology modifiers. Each modifier is defined by a list
                that includes the following elements:
                1. The path to the file that contains the modifier.
                2. The name of the function that applies the modifier.
                3. Optionally, a "hoc_string" that represents the hoc code for the modifier.
                For example, morph_modifiers could be defined as follows:

                    .. code-block::

                        morph_modifiers = [["path_to_module",
                                            "name_of_function",
                                            "hoc_string"], ...].

                If the "hoc_string" is not provided, the system will search within
                the specified module for a string that matches the function name appended
                with "_hoc".
                If ``None``, the default modifier will replace the axon with a tappered axon
                initial segment (replace_axon_with_taper). If you do not wish to use any modifier,
                set the present argument to ``[]``.
                If ``["bluepyopt_replace_axon"]``, the replace_axon function from
                bluepyopt.ephys.morphologies.NrnFileMorphology will be used
                and no other morph modifiers will be used.
            extra_mech_ids (list of 2-d tuples): extra nexus ids and types to add to
                related nexus ids. Must have shape:

                    .. code-block::

                        extra_mech_ids = [
                            ("id1", "type1"),
                            ("id2", "type2"),
                            ...
                        ]
        """

        if isinstance(parameters, dict):
            self.parameters = [ParameterConfiguration(**parameters)]
        elif isinstance(parameters, list):
            self.parameters = [ParameterConfiguration(**p) for p in parameters]
        else:
            self.parameters = []

        if isinstance(mechanisms, dict):
            self.mechanisms = [MechanismConfiguration(**mechanisms)]
        elif isinstance(mechanisms, list):
            self.mechanisms = [MechanismConfiguration(**m) for m in mechanisms]
        else:
            self.mechanisms = []

        if not distributions:
            self.distributions = []
        elif isinstance(distributions, dict):
            self.distributions = [DistributionConfiguration(**distributions)]
        elif isinstance(distributions, list):
            if isinstance(distributions[0], dict):
                self.distributions = [DistributionConfiguration(**d) for d in distributions]
            elif isinstance(distributions[0], DistributionConfiguration):
                self.distributions = distributions

        for dist in self.distributions:
            if dist.name == "uniform":
                break
        else:
            self.distributions.append(DistributionConfiguration(name="uniform"))

        self.morphology = MorphologyConfiguration(**morphology) if morphology else None

        # TODO: actually use this:
        self.mapping_multilocation = None

        self.available_mechanisms = available_mechanisms
        self.available_morphologies = available_morphologies

        self.morph_modifiers = morph_modifiers
        self.extra_mech_ids = extra_mech_ids

    @property
    def mechanism_names(self):
        """Returns the names of all the mechanisms used in the model"""

        return {m.name for m in self.mechanisms}

    @property
    def distribution_names(self):
        """Returns the names of all the distributions registered"""

        return {d.name for d in self.distributions}

    @property
    def used_distribution_names(self):
        """Returns the names of all the distributions used in the model"""

        return {p.distribution for p in self.parameters}

    @staticmethod
    def _format_locations(locations):
        """ """

        if locations is None:
            return []
        if isinstance(locations, str):
            return [locations]

        return locations

    def init_from_dict(self, configuration_dict, morphology, auto_mechanism=False):
        """Instantiate the object from its dictionary form"""

        if "distributions" in configuration_dict:
            # empty the list of distributions
            if self.distributions and configuration_dict["distributions"]:
                # remove default uniform distribution if added at
                # https://github.com/BlueBrain/BluePyEModel/blob/
                # 15b8dd2824453a8bf097b2ede13dd7ecf5d07d05/bluepyemodel/access_point/local.py#L399
                logger.warning("Removing %s pre-existing distribution(s)", len(self.distributions))
                self.distributions = []
            for distribution in configuration_dict["distributions"]:
                self.add_distribution(
                    distribution["name"],
                    distribution["function"],
                    distribution.get("parameters", None),
                    distribution.get("soma_ref_location", None),
                )

        if "parameters" in configuration_dict:
            for param in configuration_dict["parameters"]:
                self.add_parameter(
                    param["name"],
                    param["location"],
                    param["value"],
                    param.get("mechanism", None),
                    param.get("dist", None),
                    auto_mechanism=auto_mechanism,
                )

        if "mechanisms" in configuration_dict:
            for mechanism in configuration_dict["mechanisms"]:
                self.add_mechanism(
                    mechanism_name=mechanism["name"],
                    locations=mechanism["location"],
                    stochastic=mechanism.get("stochastic", None),
                    version=mechanism.get("version", None),
                    temperature=mechanism.get("temperature", None),
                    ljp_corrected=mechanism.get("ljp_corrected", None),
                    id=mechanism.get("id", None),
                )

        morphology_params = {**configuration_dict["morphology"], **morphology}
        self.morphology = MorphologyConfiguration(**morphology_params)

    def init_from_legacy_dict(self, parameters, morphology):
        """Instantiate the object from its legacy dictionary form"""

        set_mechanisms = []
        for loc in parameters["mechanisms"]:
            set_mechanisms += parameters["mechanisms"][loc]["mech"]
        set_mechanisms = set(set_mechanisms)

        for distr_name, distr in parameters["distributions"].items():
            self.add_distribution(
                distr_name,
                distr["fun"],
                distr.get("parameters", None),
                distr.get("soma_ref_location", None),
            )

        for location in parameters["parameters"]:
            if location == "__comment":
                continue

            for param in parameters["parameters"][location]:
                mechanism = None

                if (
                    param["name"] not in global_parameters
                    and "distribution" not in location
                    and "_ion" not in param["name"]
                ):
                    mechanism = next((m for m in set_mechanisms if m in param["name"]), None)
                    if mechanism is None:
                        raise ValueError(
                            f"Could not find mechanism associated to parameter {param['name']}"
                        )

                self.add_parameter(
                    param["name"],
                    location,
                    param["val"],
                    mechanism,
                    param.get("dist", None),
                )

        for location in parameters["mechanisms"]:
            for mech in parameters["mechanisms"][location]["mech"]:
                self.add_mechanism(mech, location)

        self.morphology = MorphologyConfiguration(**morphology)

    def add_distribution(
        self,
        distribution_name,
        function,
        parameters=None,
        morphology_dependent_parameters=None,
        soma_ref_location=0.5,
    ):
        """Add a channel distribution to the configuration

        Args:
            distribution_name (str): name of the distribution.
            function (str): python function of the distribution as a string. Will be executed
                using the python "eval" method.
            parameters (list of str): names of the parameters that parametrize the above function
                (no need to include the parameter "distance"). (Optional).
            morphology_dependent_parameters (list of str):
            soma_ref_location (float): location along the soma used as origin
                from which to compute the distances. Expressed as a fraction
                (between 0.0 and 1.0). (Optional).
        """

        tmp_distribution = DistributionConfiguration(
            name=distribution_name,
            function=function,
            parameters=parameters,
            morphology_dependent_parameters=morphology_dependent_parameters,
            soma_ref_location=soma_ref_location,
        )

        if any(tmp_distribution.name == d.name for d in self.distributions):
            logger.warning("Distribution %s already exists", tmp_distribution.name)

        self.distributions.append(tmp_distribution)

    def add_parameter(
        self,
        parameter_name,
        locations,
        value,
        mechanism=None,
        distribution_name=None,
        stochastic=None,
        auto_mechanism=False,
    ):
        """Add a parameter to the configuration

        Args:
            parameter_name (str): name of the parameter. If related to a mechanisms, has to match
                the name of the parameter in the mod file.
            locations (str or list of str): sections of the neuron on which these parameters
                will be instantiated.
            value (float or list of two floats): if float, set the value of the parameter. If list
                of two floats, sets the upper and lower bound between which the parameter will
                be optimised.
            mechanism (name): name of the mechanism to which the parameter relates (optional).
            distribution_name (str): name of the distribution followed by the parameter.
                Distributions have to be added before adding the parameters that uses them.
            stochastic (bool): Can the mechanisms to which the parameter relates behave
                stochastically (optional).
            auto_mechanism (bool): if True, will automatically add the mechanism associated to
                the parameter. Attention! auto_mechanism cannot handle
                mechanism version, temperature and ljp correction.
        """

        if not locations:
            raise ValueError(
                "Cannot add a parameter without specifying a location. If "
                "global parameter, put 'global'."
            )

        locations = self._format_locations(locations)

        if distribution_name is None or distribution_name == "constant":
            distribution_name = "uniform"

        if distribution_name not in self.distribution_names:
            raise ValueError(
                f"No distribution of name {distribution_name} in the configuration."
                " Please register your distributions first."
            )

        for loc in locations:
            tmp_param = ParameterConfiguration(
                name=parameter_name,
                location=loc,
                value=value,
                distribution=distribution_name,
                mechanism=mechanism,
            )

            if any(p == tmp_param for p in self.parameters):
                logger.warning(
                    "Parameter %s is already at location %s or 'all'. %s",
                    parameter_name,
                    loc,
                    "Only first occurence will be used.",
                )
            else:
                self.parameters.append(tmp_param)

            if mechanism and auto_mechanism:
                self.add_mechanism(mechanism, loc, stochastic=stochastic)

    def is_mechanism_available(
        self, mechanism_name, version=None, temperature=None, ljp_corrected=None
    ):
        """Is the mechanism part of the mechanisms available"""

        if self.available_mechanisms is not None:
            for mech in self.available_mechanisms:
                if mechanism_name in mech.name:
                    if version is None or (version is not None and version == mech.version):
                        if temp_ljp_check(temperature, ljp_corrected, mech):
                            return True

            return False

        return True

    def add_mechanism(
        self,
        mechanism_name,
        locations,
        stochastic=None,
        version=None,
        temperature=None,
        ljp_corrected=None,
        auto_parameter=False,
        id=None,
    ):
        """Add a mechanism to the configuration. This function should rarely be called directly as
        mechanisms are added automatically when using add_parameters. But it might be needed if a
        mechanism is not associated to any parameters.

        Args:
            mechanism_name (str): name of the mechanism.
            locations (str or list of str): sections of the neuron on which this mechanism
                will be instantiated.
            stochastic (bool): Can the mechanisms behave stochastically (optional).
            version (str): version id of the mod file.
            temperature (int): temperature associated with the mechanism if any
            ljp_corrected (bool): whether the mechanims is ljp corrected
            auto_parameter (bool): if True, will automatically add the parameters of the mechanism
            if they are known.
            id (str): Nexus ID of the mechanism.
        """

        locations = self._format_locations(locations)

        if mechanism_name not in NEURON_BUILTIN_MECHANISMS and not self.is_mechanism_available(
            mechanism_name, version, temperature, ljp_corrected
        ):
            raise ValueError(
                f"You are trying to add mechanism {mechanism_name} (version {version}) "
                f"(with temperature {temperature} C) (ljp_corrected is {ljp_corrected}) "
                "but it is not available on Nexus or local."
            )

        for loc in locations:
            if self.available_mechanisms and mechanism_name not in NEURON_BUILTIN_MECHANISMS:
                mechanism_parameters = next(
                    m.parameters
                    for m in self.available_mechanisms
                    if (m.name == mechanism_name and temp_ljp_check(temperature, ljp_corrected, m))
                )
            else:
                mechanism_parameters = None

            tmp_mechanism = MechanismConfiguration(
                name=mechanism_name,
                location=loc,
                stochastic=stochastic,
                version=version,
                temperature=temperature,
                ljp_corrected=ljp_corrected,
                parameters=mechanism_parameters,
                id=id,
            )

            # Check if mech is not already part of the configuration
            for m in self.mechanisms:
                if (
                    m.name == mechanism_name
                    and m.location == loc
                    and temp_ljp_check(temperature, ljp_corrected, m)
                ):
                    return

            # Handle the case where the new mech is a key of the multilocation map
            if loc in multiloc_map:
                tmp_mechanisms = []
                for m in self.mechanisms:
                    if not (
                        m.name == mechanism_name
                        and m.location in multiloc_map[loc]
                        and temp_ljp_check(temperature, ljp_corrected, m)
                    ):
                        tmp_mechanisms.append(m)
                self.mechanisms = tmp_mechanisms + [tmp_mechanism]

            # Handle the case where the new mech is a value of the multilocation map
            else:
                for m in self.mechanisms:
                    if m.name == tmp_mechanism.name and temp_ljp_check(
                        temperature, ljp_corrected, m
                    ):
                        if m.location in multiloc_map and loc in multiloc_map[m.location]:
                            return

                self.mechanisms.append(tmp_mechanism)

            if auto_parameter:
                for mechanism_parameter in tmp_mechanism.parameters:
                    self.add_parameter(
                        parameter_name=mechanism_parameter,
                        locations=[tmp_mechanism.location],
                        value=None,
                        mechanism=tmp_mechanism.name,
                        distribution_name="uniform",
                        auto_mechanism=False,
                    )

    def set_parameter_distribution(self, parameter_name, location, distribution_name):
        """Set the distribution of a parameter"""

        if not location:
            raise ValueError("Cannot set a parameter's distribution without specifying a location.")
        locations = self._format_locations(location)

        for loc in locations:
            for p in self.parameters:
                if p.name == parameter_name and p.location == loc:
                    p.distribution = distribution_name

    def set_parameter_value(self, parameter_name, location, value):
        """Set the value of a parameter"""

        if not location:
            raise ValueError("Cannot set a parameter's distribution without specifying a location.")
        locations = self._format_locations(location)

        for loc in locations:
            for p in self.parameters:
                if p.name == parameter_name and p.location == loc:
                    p.value = value

    def select_morphology(
        self,
        morphology_name=None,
        morphology_path=None,
        morphology_format=None,
        seclist_names=None,
        secarray_names=None,
        section_index=None,
    ):
        """Select the morphology on which the neuron model will be based. Its name has to be part
        of the availabel morphologies.

        Args:
            morphology_name (str): name of the morphology. Optional if morphology_path is informed.
            morphology_path (str): path to the morphology file. If morphology_name is informed, has
                to match the stem of the morphology_path. Optional if morphology_name is informed.
            morphology_format (str): format of the morphology (asc or swc). If morphology_path is
                informed, has to match its suffix. Optional if morphology_format is informed.
            seclist_names (list): Names of the lists of sections (['somatic', ...]) (optional).
            secarray_names (list): names of the sections (['soma', ...]) (optional).
            section_index (int): index to a specific section, used for non-somatic
                recordings (optional).
        """

        if morphology_name not in self.available_morphologies:
            raise ValueError(
                "The morphology you want to use is not available in Nexus or in your "
                "morphology directory"
            )

        self.morphology = MorphologyConfiguration(
            name=morphology_name,
            path=morphology_path,
            format=morphology_format,
            seclist_names=seclist_names,
            secarray_names=secarray_names,
            section_index=section_index,
        )

    def remove_parameter(self, parameter_name, locations=None):
        """Remove a parameter from the configuration. If locations is None or [], the whole
        parameter will be removed. WARNING: that does not remove automatically the mechanism
        which might be still use by other parameter"""

        locations = self._format_locations(locations)

        if locations:
            self.parameters = [
                p
                for p in self.parameters
                if p.name != parameter_name or p.location not in locations
            ]
        else:
            self.parameters = [p for p in self.parameters if p.name != parameter_name]

    def remove_mechanism(self, mechanism_name, locations=None):
        """Remove a mechanism from the configuration and all the associated parameters"""

        locations = self._format_locations(locations)

        if locations:
            self.mechanisms = [
                m
                for m in self.mechanisms
                if m.name != mechanism_name or m.location not in locations
            ]
            self.parameters = [
                p
                for p in self.parameters
                if p.mechanism != mechanism_name or p.location not in locations
            ]
        else:
            self.mechanisms = [m for m in self.mechanisms if m.name != mechanism_name]
            self.parameters = [p for p in self.parameters if p.mechanism != mechanism_name]

    def get_related_nexus_ids(self):
        uses = []
        mechs_ids = set()
        if self.morphology.id:
            uses.append({"id": self.morphology.id, "type": "NeuronMorphology"})

        for m in self.mechanisms:
            if m.id and m.id not in mechs_ids:
                uses.append({"id": m.id, "type": "SubCellularModelScript"})
                mechs_ids.add(m.id)

        if self.extra_mech_ids is not None:
            for mech_id, mech_type in self.extra_mech_ids:
                if mech_id not in mechs_ids:
                    uses.append({"id": mech_id, "type": mech_type})
                    mechs_ids.add(mech_id)

        return {"uses": uses}

    def as_dict(self):
        """Returns the configuration as dict of parameters, mechanisms and
        a list of mechanism names"""

        return {
            "mechanisms": [m.as_dict() for m in self.mechanisms],
            "distributions": [
                d.as_dict() for d in self.distributions if d.name in self.used_distribution_names
            ],
            "parameters": [p.as_dict() for p in self.parameters],
            "morphology": self.morphology.as_dict(),
            "morph_modifiers": self.morph_modifiers,
        }

    def __str__(self):
        """String representation"""

        str_form = "Model Configuration:\n\n"

        str_form += "Mechanisms:\n"
        for m in self.mechanisms:
            str_form += f"   {m.as_dict()}\n"

        str_form += "Distributions:\n"
        for d in self.distributions:
            if d.name in self.used_distribution_names:
                str_form += f"   {d.as_dict()}\n"

        str_form += "Parameters:\n"
        for p in self.parameters:
            str_form += f"   {p.as_dict()}\n"

        str_form += "Morphology:\n"
        str_form += f"   {self.morphology.as_dict()}\n"

        str_form += "Morphology modifiers:\n"
        if isinstance(self.morph_modifiers, list):
            for morph_modifier in self.morph_modifiers:
                str_form += f"   {morph_modifier}\n"
        else:
            str_form += f"   {self.morph_modifiers}\n"

        return str_form



"""Morphology Configuration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pathlib


class MorphologyConfiguration:
    """Morphology configuration"""

    def __init__(
        self,
        name=None,
        path=None,
        format=None,
        seclist_names=None,
        secarray_names=None,
        section_index=None,
        id=None,
    ):
        """Init.

        Args:
            name (str): name of the morphology. If None, the file name from the path will
                be used.
            path (str): path to the morphology file.
            format (str): format of the morphology file, as to be 'asc' or 'swc'. If None,
                the extension of the path will be used.
            seclist_names (list): Optional. Names of the lists of sections
                (e.g: ['somatic', ...]).
            secarray_names (list): Optional. Names of the sections (e.g: ['soma', ...]).
            section_index (list): Optional. Index to a specific section, used for
                non-somatic recordings.
            id (str): Optional. Nexus ID of the morphology.
        """

        self.path = None
        if path:
            self.path = str(path)

        if name:
            self.name = name
        elif path:
            self.name = pathlib.Path(self.path).stem
        else:
            raise TypeError("name or path has to be informed")

        self.format = None
        if format:
            if self.path:
                if format.lower() != path[-3:].lower():
                    raise ValueError("The format does not match the morphology file")
            self.format = format
        elif self.path:
            self.format = path[-3:]

        if self.format and self.format.lower() not in ["asc", "swc"]:
            raise ValueError("The format of the morphology has to be 'asc' or 'swc'.")

        self.seclist_names = seclist_names
        self.secarray_names = secarray_names
        self.section_index = section_index

        self.id = id

    def as_dict(self):
        return {
            "name": self.name,
            "format": self.format,
            "path": self.path,
            "seclist_names": self.seclist_names,
            "secarray_names": self.secarray_names,
            "section_index": self.section_index,
            "id": self.id,
        }



"""Utils"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

logger = logging.getLogger("__main__")


def temp_ljp_check(temperature, ljp_corrected, mech):
    """Returns True if mech has corresponding temperature and ljp correction

    Args:
        temperature (int): temperature associated with the mechanism if any
        ljp_corrected (bool): whether the mechanism is ljp corrected
        mech (MechanismConfiguration): mechanism
    """
    if temperature is None or temperature == mech.temperature:
        if ljp_corrected is None or ljp_corrected is mech.ljp_corrected:
            return True
    return False



"""sAHP stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class sAHP(BPEM_stimulus):
    """sAHP current stimulus

    The long step (here amp) is usually fixed at 40% of rheobase, and the short step (here amp2)
    can usually vary from 150% to 300% of rheobase.

    .. code-block:: none

           holdi       holdi+long_amp      holdi+amp       holdi+long_amp          holdi
             :                :                :                 :                   :
             :                :          ______________          :                   :
             :                :         |              |         :                   :
             :      ____________________|              |____________________         :
             :     |                    ^              ^                    |        :
             :     |                    :              :                    |        :
        |__________|                    :              :                    |__________________
        ^          ^                    :              :                    ^                  ^
        :          :                    :              :                    :                  :
        :          :                    :              :                    :                  :
        t=0        delay                tmid           tmid2                toff     totduration

    """

    name = "sAHP"

    def __init__(self, location, **kwargs):
        """Constructor
        Args:
            location(Location): location of stimulus
        """

        self.amp = kwargs.get("amp", None)
        self.amp_rel = kwargs.get("thresh_perc", None)

        self.long_amp = kwargs.get("long_amp", None)
        self.long_amp_rel = kwargs.get("long_amp_rel", None)

        if self.amp is None and self.amp_rel is None:
            raise TypeError(f"In stimulus {self.name}, amp and thresh_perc cannot be both None.")

        if self.long_amp is None and self.long_amp_rel is None:
            raise TypeError(
                f"In stimulus {self.name}, long_amp and long_amp_rel cannot be both None."
            )

        self.holding_current = kwargs.get("holding_current", None)
        self.threshold_current = None

        self.delay = kwargs.get("delay", 250.0)
        self.tmid = kwargs.get("tmid", 500.0)
        self.tmid2 = kwargs.get("tmid2", 725.0)
        self.toff = kwargs.get("toff", 1175.0)
        self.total_duration = kwargs.get("totduration", 1425.0)

        super().__init__(
            location=location,
        )

    @property
    def stim_start(self):
        return self.delay

    @property
    def stim_end(self):
        return self.toff

    @property
    def amplitude(self):
        if self.amp_rel is None or self.threshold_current is None:
            return self.amp
        return self.threshold_current * (float(self.amp_rel) / 100.0)

    @property
    def long_amplitude(self):
        if self.long_amp_rel is None or self.threshold_current is None:
            return self.long_amp
        return self.threshold_current * (float(self.long_amp_rel) / 100.0)

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        holding_current = self.holding_current if self.holding_current is not None else 0

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()

        self.time_vec.append(0.0)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.delay)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.delay)
        self.current_vec.append(holding_current + self.long_amplitude)

        self.time_vec.append(self.tmid)
        self.current_vec.append(holding_current + self.long_amplitude)

        self.time_vec.append(self.tmid)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(self.tmid2)
        self.current_vec.append(holding_current + self.amplitude)

        self.time_vec.append(self.tmid2)
        self.current_vec.append(holding_current + self.long_amplitude)

        self.time_vec.append(self.toff)
        self.current_vec.append(holding_current + self.long_amplitude)

        self.time_vec.append(self.toff)
        self.current_vec.append(holding_current)

        self.time_vec.append(self.total_duration)
        self.current_vec.append(holding_current)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )

    def generate(self, dt=0.1):
        """Return current time series

        WARNING: do not offset ! This is on-top of a holding stimulus."""
        holding_current = self.holding_current if self.holding_current is not None else 0

        t = numpy.arange(0.0, self.total_duration, dt)
        current = numpy.full(t.shape, holding_current, dtype="float64")

        ton = int(self.delay / dt)
        tmid = int(self.tmid / dt)
        tmid2 = int(self.tmid2 / dt)
        toff = int(self.toff / dt)

        current[ton:tmid] += self.long_amplitude
        current[tmid2:toff] += self.long_amplitude
        current[tmid:tmid2] += self.amplitude

        return t, current



"""Target"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

logger = logging.getLogger(__name__)


class Target:
    """Describes an extraction (or optimisation) target"""

    def __init__(
        self,
        efeature,
        protocol,
        amplitude,
        tolerance,
        efeature_name=None,
        efel_settings=None,
        weight=1.0,
    ):
        """Constructor

        Args:
            efeature (str): name of the eFeature in the eFEL library
                (ex: 'AP1_peak')
            protocol (str): name of the recording from which the efeature
                should be computed
            amplitude (float): amplitude of the current stimuli for which the
                efeature should be computed (expressed as a percentage of the
                threshold amplitude (rheobase))
            tolerance (float): tolerance around the target amplitude in which
                an experimental recording will be seen as a hit during
                efeatures extraction (expressed as a percentage of the
                threshold amplitude (rheobase))
            efeature_name (str): given name for this specific target. Can be different
                from the efel efeature name.
            efel_settings (dict): target specific efel settings.
            weight (float): weight of the efeature.
                Basically multiplies the score of the efeature by this value.
        """

        self.efeature = efeature
        self.protocol = protocol

        self.amplitude = amplitude
        self.tolerance = tolerance

        self.efeature_name = efeature_name

        self.weight = weight

        if efel_settings is None:
            self.efel_settings = {"strict_stiminterval": True}
        else:
            self.efel_settings = efel_settings

    def as_dict(self):
        return vars(self)



""" ICSELECTOR
Select a set of NEURON mechanisms, parameters and bounds from corresponding
genes. Gene names can be selected from a file mapping genes to different
ME-T types. Corresponding channels are selected from a file mapping channels
and parameters to genes.
"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import json
import logging

import numpy as np
import pandas as pd

from .modules.configuration import Configuration
from .modules.gene_selector import GeneSelector
from .modules.model_selector import ModelSelector


class ICSelector:
    """Selects NEURON mechanisms, with parameters, value bounds, and
    distributions. Uses reference files for MET-type to gene mapping,
    gene to channel mapping, and channel to parameter mapping."""

    def __init__(self, ic_map_path, gene_map_path, mode="mixed", status="latest"):
        """
        Args:
            ic_map_path (str): path to ic_mapping file (.json)
            gene_map_path (str): path to gene mapping file (.csv)
            mode (str): types of ion channel model to select.
                Options are 'generic', 'genetic', and 'mixed' (default).
            status (str): model status. Options are 'stable' and 'latest'.
        """

        # === Load mapping files
        gene_map = pd.read_csv(gene_map_path, index_col=[0, 1, 2])
        self._gene_selector = GeneSelector(gene_map)

        with open(ic_map_path, mode="r", encoding="utf-8") as fid:
            ic_map = json.load(fid)
        self._misc_parameters = ic_map["misc_parameters"]
        self._gene_to_ic = ic_map["genes"]
        self._distributions = ic_map["distributions"]
        self._model_selector = ModelSelector(ic_map)

        self.set_status(status)
        self.set_mode(mode)

    def set_status(self, status):
        """Select model types.

        Args:
            status (str): 'stable' or 'latest'
        """
        if status in ["stable", "latest"]:
            self._model_selector.status = status

    def set_mode(self, mode):
        """Select configuration mode.

        Args:
            mode (str): 'generic', 'genetic', or 'mixed'
        """

        if mode in ["genetic", "generic", "mixed"]:
            self._model_selector.mode = mode

    def _get_channel_name(self, gene_name):
        """Get channel name from gene.

        Args:
            gene_name (str): gene name

        Returns:
            (str): channel name
        """
        if gene_name in self._gene_to_ic:
            return self._gene_to_ic[gene_name]
        return None

    def _get_mech_from_gene(self, gene_name):
        """Get channel mechanism from gene.

        Args:
            gene_name (str): gene name

        Returns:
            (Mechanism): channel mechanism
        """

        channel_name = self._get_channel_name(gene_name)
        if channel_name:
            return self._model_selector.get(channel_name)
        return None

    def __set_parameters_from_gene(self, gene_name):
        """Set mechanism fields corresponding to gene.
        Args:
            gene_name (str): gene name
        """

        info = self._gene_selector.get(gene_name)
        channel_name = info["channel"]
        mech = self.get(channel_name)
        if mech:
            mech.set_from_gene_info(info)
        else:
            logging.warning("Could not determine mechanism for gene %s", gene_name)

    def _set_parameters_from_ttype(self):
        """Copy parameters from selected genes to mechanims."""

        for gene_name in self._gene_selector.selected_genes:
            self.__set_parameters_from_gene(gene_name)

    def _select_genes_from_ttype(self, key_words):
        """Select genes from key words.
        Args:
            key_words (list [str]): list of keys to select genes
        """

        # === Get genes from gene mapping file
        logging.info("\n===============\nGenes Selection\n===============")
        genes = self._gene_selector.select_from_ttype(key_words)
        logging.info(str(self._gene_selector))

        # === Map genes to channels
        for gene, info in genes.items():
            if gene in self._gene_to_ic:
                info["channel"] = self._gene_to_ic[gene]

    def _select_mechanisms_from_ttype(self):
        """Select mechanisms from previously selected genes."""

        logging.info("\n==================\nChannels Selection\n==================")
        for gene_name in self._gene_selector.selected_genes:
            self.select(gene_name)
        logging.info(str(self._model_selector))

    def get(self, name):
        """Get mechanism corresponding to gene or channel.

        Args:
            name (str): name of a gene or channel.

        Returns:
            mech (Mechanism): selected mechanism
        """

        # Try first to select as channel name
        mech = self._model_selector.get(name)
        if not mech:
            # If that fails, try as gene name
            mech = self._get_mech_from_gene(name)
        return mech

    def select(self, name):
        """Select mechanism for inclusion in configuration.

        Args:
            name (str): name of a gene or channel.
        """

        selected = self._model_selector.select(name)
        if not selected:
            channel_name = self._get_channel_name(name)
            selected = self._model_selector.select(channel_name)

    def get_gene_mapping(self):
        """Get a dict of all selected genes and mapped channels.
        Returns:
            genes (dict): all selected genes and mapped channels
        """

        mechs = self.get_mechanisms()
        genes = self._gene_selector.selected_genes
        genes = {k: v for k, v in genes.items() if not v["channel"] == "n/a"}
        for k, v in genes.items():
            channel = v["channel"]
            for a, b in mechs.items():
                if channel in b["_mapped_from"]:
                    v["mapped_to"] = a
        return genes

    def get_mechanisms(self, selected_only=True):
        """Get all available mechanisms from the icmapping file.

        Args:
            selected_only (bool): flag to get only selected channels

        Returns:
            mechs (dict [Mechanism]): mechanisms with all associated info
                fields
        """

        mechs = self._model_selector.get_mechanisms(selected_only)
        mechs = {k: v.asdict() for k, v in mechs.items()}
        return mechs

    def get_selected_cell_types(self):
        """Get met types selected from the gene mapping table.

        Returns:
            (list): all selected cell types
        """

        return self._gene_selector.selected_met_types

    def _get_cell_config(self):
        """Generate cell model configuration from selected mechanisms.

        Returns:
            config (Configuration): set of mechanism parameters per compartment
        """

        mechanisms = self._model_selector.get_mechanisms()
        config = Configuration()
        for mech in mechanisms.values():
            config.add_from_mechanism(mech)

        logging.info("\n=======================\nParameter Configuration\n=======================")

        # === Set additional parameters
        misc = self._misc_parameters
        for name, setting in misc.items():

            if "local" in setting:
                # Location to be obtained from mechanism
                locations = []
                for mech in mechanisms.values():
                    model = mech.model
                    if "read" in model:
                        reads = model["read"]
                        if name in reads:
                            locs = mech.distribution.get()
                            locations.extend(locs.keys())
                for loc in np.unique(locations):
                    config.add_parameter(name, loc, setting["local"])
            else:
                for loc, value in setting.items():
                    config.add_parameter(name, loc, value)

        logging.info(str(config))
        return config

    def get_cell_config_from_ttype(self, key_words):
        """Get all information related to cell model configuration from
        mechanisms selected based on genetic expression profiles.

        Args:
            key_words: keys to select MET-types

        Returns:
            parameters (list [dict,]): mechanism parameters per compartment
            mechanisms (list [dict,]): mechanisms per compartment
            distributions (list [dict,]): definitions of subcellular distributions
            nexus_keys (list [dict,]): name and modelid for each mechanism
        """

        # Perform selections
        self._select_genes_from_ttype(key_words)
        self._set_parameters_from_ttype()
        self._select_mechanisms_from_ttype()

        # Get configuration
        mechs = self._model_selector.get_mechanisms()
        config = self._get_cell_config()

        # Get configuration parameters
        parameters = config.get_parameters()
        mechanisms = config.get_mechanisms()
        distributions = config.get_distributions()

        # Define distributions
        distr_funcs = self._distributions
        for distr in distributions:
            distr_name = distr["name"]
            if not distr_name == "uniform":
                if distr_name in distr_funcs:
                    d = distr_funcs[distr_name]
                    for k, v in d.items():
                        distr[k] = v
                else:
                    logging.warning("Unknown distribution '%s'.", distr_name)

        # Get Nexus keys
        nexus_keys = [v.nexus for k, v in mechs.items() if v.nexus]

        return parameters, mechanisms, distributions, nexus_keys



"""ProbAMPANMDA_EMS class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from bluepyemodel.ecode.stimulus import BPEM_stimulus

logger = logging.getLogger(__name__)


class ProbAMPANMDA_EMS(BPEM_stimulus):
    """ProbAMPANMDA_EMS synapse current injection"""

    name = "ProbAMPANMDA_EMS"

    def __init__(self, location=None, **kwargs):
        """Constructor

        Args:
            location (Location): stimulus Location
            syn_weight (float): synaptic weight
            syn_delay (float) synaptic delay (ms)
            total_duration (float): total duration (ms)
            use (float): synapse Use
            stimfreq (float): stimulus frequency (Hz)
            number (int): number of synaptic spikes
            noise (float): range 0 to 1. Fractional randomness
            netcon_thres (float): NetCon threshold (mV)
            netcon_delay (float): NetCon delay (ms)
            netcon_weight (float): NetCon weight
        """
        self.syn_weight = kwargs.get("syn_weight", None)
        self.syn_delay = kwargs.get("syn_delay", None)
        self.total_duration = kwargs.get("totduration", 100.0)
        self.use = kwargs.get("use", 1.0)
        self.stimfreq = kwargs.get("stimfreq", 70.0)
        self.number = kwargs.get("number", 1)
        self.noise = kwargs.get("noise", 0)
        self.netcon_thres = kwargs.get("netcon_thres", 10)
        self.netcon_delay = kwargs.get("netcon_delay", 0)
        self.netcon_weight = kwargs.get("netcon_weight", 700)
        self.synapse = None
        self.netstim = None
        self.netcon = None

        if self.syn_weight is None or self.syn_delay is None:
            raise TypeError(
                f"syn_weight and syn_delay should be specified for stimulus {self.name}"
            )

        super().__init__(
            location=location,
        )

    @property
    def stim_start(self):
        return self.syn_delay

    @property
    def stim_end(self):
        return self.total_duration

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        icomp = self.location.instantiate(sim=sim, icell=icell)
        logger.debug(
            "Adding synaptic excitatory stimulus to %s with delay %f and weight %f",
            str(self.location),
            self.syn_delay,
            self.syn_weight,
        )

        self.synapse = sim.neuron.h.ProbAMPANMDA_EMS(icomp.x, sec=icomp.sec)
        self.synapse.Use = self.use

        self.netstim = sim.neuron.h.NetStim(sec=icomp.sec)
        self.netstim.interval = 1000 / self.stimfreq
        self.netstim.number = self.number
        self.netstim.start = self.syn_delay
        self.netstim.noise = self.noise
        self.netcon = sim.neuron.h.NetCon(
            self.netstim,
            self.synapse,
            self.netcon_thres,
            self.netcon_delay,
            self.netcon_weight,
            sec=icomp.sec,
        )
        self.netcon.weight[0] = self.syn_weight

    def destroy(self, sim=None):
        """Destroy stimulus"""
        super().destroy(sim=sim)

        self.synapse = None
        self.netstim = None
        self.netcon = None



#!/usr/bin/env python

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from setuptools import find_packages, setup

# Read the contents of the README file
with open("README.rst", encoding="utf-8") as f:
    README = f.read()


EXTRA_LUIGI = [
    "luigi>=3.0",
    "luigi-tools>=0.0.12",
]

EXTRA_TEST = ["pytest>=6.2", "dictdiffer>=0.8"]

EXTRA_DOC = [
    "graphviz",
    "sphinx",
    "sphinx-bluebrain-theme",
]


setup(
    name="bluepyemodel",
    use_scm_version={
        "version_scheme": "python-simplified-semver",
        "local_scheme": "no-local-version",
    },
    setup_requires=["setuptools_scm"],
    author="Blue Brain Project, EPFL",
    author_email="",
    description="Blue Brain Python Electrical Modeling Pipeline",
    long_description=README,
    long_description_content_type="text/x-rst",
    license="Apache-2.0",
    python_requires=">=3.9",
    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "ipyparallel>=6.3",
        "tqdm",
        "pyyaml",
        "gitpython",
        "bluepyopt>=1.14.10",
        "bluepyefe>=2.2.0",
        "neurom>=3.0",
        "efel>=5.5.5",
        "configparser",
        "neuron>=8.0",
        "morph_tool>=2.8",
        "morphio",
        "fasteners>=0.16",
        "jinja2>=3.0.3",
        "currentscape>=0.0.11",
    ],
    extras_require={
        "luigi": EXTRA_LUIGI,
        "all": EXTRA_LUIGI + EXTRA_TEST,
        "docs": EXTRA_DOC + EXTRA_LUIGI,
        "test": EXTRA_TEST,
    },
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    keywords=[
        "computational neuroscience",
        "simulation",
        "analysis",
        "parameters",
        "Blue Brain Project",
    ],
    url="https://github.com/BlueBrain/BluePyEModel",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
)



"""Parameter Configuration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class ParameterConfiguration:
    """Contains all the information related to the definition and configuration of a parameter"""

    def __init__(self, name, location, value, distribution="uniform", mechanism=None):
        """Init

        Args:
            name (str): name of the parameter. If related to a mechanisms, has to match
                the name of the parameter in the mod file.
            location (str): section of the neuron on which the parameter will be instantiated.
            value (float or list of two floats): if float, set the value of the parameter. If list
                of two floats, sets the upper and lower bound between which the parameter will
                be optimised.
            distribution (str): name of the distribution followed by the parameter (optional).
            mechanism (name): name of the mechanism to which the parameter relates (optional).
        """

        self.name = name
        self.location = location

        self.value = value
        if isinstance(self.value, tuple):
            self.value = list(self.value)
        if isinstance(self.value, list) and len(self.value) == 1:
            self.value = self.value[0]

        self.mechanism = mechanism

        self.distribution = distribution
        if self.distribution is None:
            self.distribution = "uniform"

    @property
    def valid_value(self):
        return not (self.value is None)

    def as_dict(self):
        """ """

        param_dict = {
            "name": self.name,
            "value": self.value,
            "location": self.location,
        }

        if self.distribution and self.distribution != "uniform":
            param_dict["distribution"] = self.distribution

        if self.mechanism:
            param_dict["mechanism"] = self.mechanism

        return param_dict

    def as_legacy_dict(self):
        """ """

        param_dict = {"name": self.name, "val": self.value}

        if self.distribution and self.distribution != "uniform":
            param_dict["dist"] = self.distribution

        return param_dict

    def __eq__(self, other):
        return self.name == other.name and (
            self.name == "all" or other.location == "all" or self.location == other.location
        )



"""Luigi tool module."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import json
import logging
from abc import ABC

import luigi

try:
    from bbp_workflow.task import IPyParallelExclusive
except ImportError as exc:
    raise ImportError("The internal bbp-workflow package is required to use Luigi.") from exc
from luigi.parameter import MissingParameterException
from luigi.parameter import _no_value
from luigi_tools.task import _no_default_value

from bluepyemodel import access_point
from bluepyemodel.tasks.config import EmodelAPIConfig
from bluepyemodel.tools.multiprocessing import get_mapper

logger = logging.getLogger(__name__)


class WorkflowTask(luigi.Task):
    """Workflow task with loaded data access point."""

    backend = luigi.Parameter(default=None, config_path={"section": "parallel", "name": "backend"})

    emodel = luigi.Parameter()
    etype = luigi.Parameter(default=None)
    ttype = luigi.OptionalParameter(default=None)
    mtype = luigi.OptionalParameter(default=None)
    species = luigi.Parameter(default=None)
    brain_region = luigi.Parameter(default=None)
    iteration_tag = luigi.Parameter(default=None)

    def __init__(self, *args, **kwargs):
        """ """
        super().__init__(*args, **kwargs)

        self.access_point = access_point.get_access_point(
            EmodelAPIConfig().api,
            emodel=self.emodel,
            etype=self.etype,
            ttype=self.ttype,
            mtype=self.mtype,
            species=self.species,
            brain_region=self.brain_region,
            iteration_tag=self.iteration_tag,
            **EmodelAPIConfig().api_args
        )

    def get_mapper(self):
        """Get a mapper for parallel computations."""
        return get_mapper(self.backend)

    @staticmethod
    def check_mettypes(func):
        """Decorator to check mtype, etype and ttype presence on nexus"""

        def inner(self):
            """Inner decorator function"""
            if EmodelAPIConfig().api == "nexus":
                self.access_point.check_mettypes()
            func(self)

        return inner


class WorkflowTaskRequiringMechanisms(WorkflowTask):
    """Workflow task with data access point and download of missing mechanisms"""

    def __init__(self, *args, **kwargs):
        """ """

        super().__init__(*args, **kwargs)

        self.access_point.get_model_configuration()


class WorkflowTarget(luigi.Target, ABC):
    """Workflow target with loaded access_point."""

    def __init__(
        self, emodel, etype, ttype, mtype, species, brain_region, iteration_tag, *args, **kwargs
    ):
        """ """
        super().__init__(*args, **kwargs)

        self.access_point = access_point.get_access_point(
            EmodelAPIConfig().api,
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=species,
            brain_region=brain_region,
            iteration_tag=iteration_tag,
            **EmodelAPIConfig().api_args
        )


class WorkflowWrapperTask(WorkflowTask, luigi.WrapperTask):
    """Base wrapper class with global parameters."""


class IPyParallelTask(IPyParallelExclusive):
    """Wrapper around IPyParallelTask to get chdir param from [DEFAULT] in config."""

    chdir = luigi.configuration.get_config().get("DEFAULT", "chdir")

    def prepare_args_for_remote_script(self, attrs):
        """Prepare self.args, which is used to pass arguments to remote_script."""
        # start with '--' to separate ipython arguments from parsing arguments
        self.args = "--"  # pylint:disable=W0201

        for attr in attrs:
            if hasattr(self, attr):
                # Luigi stores lists as tuples, but json cannot load tuple
                # so here, turning tuples back into lists.
                # Also turn lists and dicts into json strings.
                if isinstance(getattr(self, attr), tuple):
                    setattr(self, attr, json.dumps(list(getattr(self, attr))))
                # luigi stores dicts as luigi.freezing.FrozenOrderedDict
                # that are not json serializable,
                # so turn them into dict, and then into json strings
                elif isinstance(getattr(self, attr), (dict, luigi.freezing.FrozenOrderedDict)):
                    setattr(self, attr, json.dumps(dict(getattr(self, attr))))

                if getattr(self, attr) is True:
                    # pylint:disable=W0201
                    self.args = " ".join([self.args, "--" + attr])
                elif getattr(self, attr) is not False and getattr(self, attr) is not None:
                    # be sure that lists and dicts are inside ' '
                    # so that argparse detect them as one argument
                    # have to change ' to '\\'' because args would already be
                    # inside quotes (') in command from bbp-workflow
                    # single quotes would mess that up
                    self.args = " ".join(  # pylint:disable=W0201
                        [
                            self.args,
                            "--" + attr,
                            "'\\''" + str(getattr(self, attr)) + "'\\''",
                        ]
                    )

        # append API-related arguments
        # api is str
        self.args += " --api_from_config " + EmodelAPIConfig().api
        # api_args is dict
        self.args += (
            " --api_args_from_config "
            + "'\\''"
            + json.dumps(dict(EmodelAPIConfig().api_args))
            + "'\\''"
        )

        # append ipyparallel profile (str)
        self.args += " --ipyparallel_profile " + self.task_id


class BoolParameterCustom(luigi.BoolParameter):
    """Class to make luigi BoolParameter compatible with luigi-tools's copy-params."""

    def task_value(self, task_name, param_name):
        value = self._get_value(task_name, param_name)
        if value == _no_value:
            raise MissingParameterException("No default specified")
        if value == _no_default_value:
            return _no_default_value
        return self.normalize(value)

    def parse(self, val):
        """
        Parses a ``bool`` from the string, matching 'true' or 'false' ignoring case.
        """
        if val == _no_default_value:
            return _no_default_value
        return super().parse(val)


class ListParameterCustom(luigi.ListParameter):
    """Class to make luigi ListParameter compatible with luigi-tools's copy-params.

    When a class that has copy-params is yielded, this class should replace luigi.ListParameter.
    """

    def parse(self, x):
        """
        Parse an individual value from the input.

        Do not parse if value is None or _no_default_value

        :param str x: the value to parse.
        :return: the parsed value.
        """
        if x is None:
            return None
        if x == _no_default_value:
            return _no_default_value
        return super().parse(x)

    def serialize(self, x):
        """
        Opposite of :py:meth:`parse`.

        Converts the value ``x`` to a string unless x is _no_default_value.

        :param x: the value to serialize.
        """
        if x == _no_default_value:
            return _no_default_value
        return super().serialize(x)



"""Methods for selecting genes associated with a given me- and t-type"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import logging

import numpy as np

from .distribution import Distribution


class GeneSelector:
    """Selects genes from Yann's gene mapping file."""

    def __init__(self, gene_map):
        """
        Args:
            gene_map (DataFrame): Array mapping genes to met-types. Rows have
            met-types, columns have genes
        """
        self._gene_map = gene_map
        self.selected_genes = {}
        self.selected_met_types = []

    @staticmethod
    def _filter(df, keys):
        """Rules for filtering the columns of Yann's gene map.

        Args:
            df (DataFrame): subset of Yann's table
            keys (list [str]): list of strings to filter columns

        Returns:
            df (DataFrame): subset of the table
        """

        # First filter on exact match
        for name in df.index.names:
            crit = df.index.get_level_values(name).isin(keys)
            # Only apply filter if exact match was found
            if not np.all(~crit):
                df = df[crit]
        # Then filter on partial match
        for key in keys:
            new_df = df.filter(regex=key, axis=0)
            # Only apply filter if partial match was found
            if not new_df.shape[0] == 0:
                df = new_df
        return df

    @staticmethod
    def _get_gene_presence(gene):
        """Determine if a gene is present or not.

        Args:
            gene (DataFrame): single column from the gene table

        Returns:
            presence (array [bool]): gene is present
        """

        gene_present = gene["presence"]
        if isinstance(gene_present, str):
            presence = np.array([int(gene_present)], dtype=int) == 1
        else:
            presence = np.array(gene_present.values, dtype=int) == 1
        return presence

    @staticmethod
    def _get_distributions(gene, presence, group_compartments):
        """Get distribution per compartment for a given gene.

        Args:
            gene (DataFrame): single column from the gene table
            presence (array [bool]): gene is present
            group_compartments (bool): use 'all' if present in all compartments

        Returns:
            distros (Distribution): gene distribution per compartment
        """

        all_comps = None
        dend = np.unique(np.array(gene.filter(regex="dend", axis=0).values[presence], dtype=str))
        axon = np.unique(np.array(gene.filter(regex="axon", axis=0).values[presence], dtype=str))
        soma = np.unique(np.array(gene.filter(regex="soma", axis=0).values[presence], dtype=str))

        if len(dend) > 1:
            logging.warning(
                "Conflicting dendritic distributions for gene %s. Selected '%s' from %s",
                gene.name,
                dend[0],
                dend,
            )
        dend = dend[0]
        if len(axon) > 1:
            logging.warning(
                "Conflicting axonal distributions for gene %s. Selected '%s' from %s",
                gene.name,
                axon[0],
                axon,
            )
        axon = axon[0]
        if len(soma) > 1:
            logging.warning(
                "Conflicting somatic distributions for gene %s. Selected '%s' from %s",
                gene.name,
                soma[0],
                soma,
            )
        soma = soma[0]

        if (dend == axon) and (axon == soma):
            all_comps = dend

        def str_has_value(s):
            return not ((s == "nan") or (s.strip() == ""))

        distros = Distribution()
        if all_comps and group_compartments:
            if str_has_value(all_comps):
                distros.all = all_comps
        else:
            distros.set_fields(somatic=soma, axonal=axon, basal=dend, apical=dend)
        return distros

    @staticmethod
    def _get_gbar_max(gene):
        """Get gbar_max for given gene.

        Args:
            gene (DataFrame): single column from the gene table

        Returns:
            gbar_max (float): gbar_max value
        """

        try:
            gbar_max = gene["g_bar_max"]
            if isinstance(gbar_max, (str, float)):
                gbar_max = np.array([float(gbar_max)])
            else:
                gbar_max = np.array([float(g) for g in gbar_max.values])
        except KeyError:
            gbar_max = np.array([0.0])
        crit = np.array([np.isnan(g) for g in gbar_max])
        gbar_max[crit] = 0.0
        gbar_max = np.unique(gbar_max)
        if len(gbar_max) > 1:
            logging.warning(
                "Conflicting g_bar_max values for gene %s. Selected '%s' from %s",
                gene.name,
                np.max(gbar_max),
                gbar_max,
            )
        return np.max(gbar_max)

    def get(self, gene_name):
        """Get info for given gene.

        Args:
            gene_name (str): name of gene

        Returns:
            (dict): gene info
        """

        if gene_name in self.selected_genes:
            return self.selected_genes[gene_name]
        raise KeyError("Gene not available for the selected ttype.")

    def select_from_ttype(self, keys=None, group_compartments=False):
        """Returns selected genes and distributions associated with provided
        key words.

        Args:
            keys (list [str]): List of keywords
            group_compartments: Option to combine compartments into groups
            (e.g. 'all', or 'alldend')

        Returns:
            selected_genes (list [dict]): list of selected genes with info
        """

        df = self._gene_map
        if isinstance(keys, str):
            keys = [keys]
        if keys:
            df = self._filter(df, keys)
        # Store result
        self.selected_met_types = np.unique([f"{v[0]} - {v[1]}" for v in df.index.values])
        df = df.droplevel([0, 1])
        # Apply filter also to genes
        if keys:
            genes = self._filter(df.T, keys)
            genes = genes.T
        else:
            genes = df

        names = genes.columns.values
        for name in names:
            gene = genes[name]
            # Check if gene is present
            presence = self._get_gene_presence(gene)
            if sum(presence) == 0:
                continue
            # Determine distributions per compartment
            distros = self._get_distributions(gene, presence, group_compartments)
            # Determine gbar_max
            gbar_max = self._get_gbar_max(gene)
            # Add item
            self.selected_genes[name] = {
                "channel": "n/a",
                "distribution": distros,
                "gbar_max": gbar_max,
            }

        return self.selected_genes

    def __str__(self):
        heading_str = "\n>>> M/E/T types <<<"
        out_str = [heading_str]
        for k in self.selected_met_types:
            out_str += [k]
        out_str += ["\n>>> Genes <<<"]
        items = self.selected_genes.items()
        n = 1
        for k, v in items:
            items_str = f"-{n}- {k}, distribution: {v['distribution']}, gbar_max: {v['gbar_max']}"
            n += 1
            out_str += [items_str]
        return "\n".join(out_str)



"""Ecode for dendrite specific protocols, such as synaptic input, dendritic steps, or BAC."""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from bluepyopt.ephys.locations import NrnSomaDistanceCompLocation
from bluepyopt.ephys.locations import NrnTrunkSomaDistanceCompLocation

from .idrest import IDrest


class DendriticStep(IDrest):
    """Step protocol on a dendrite."""

    name = "DendriticStep"

    def __init__(self, location, **kwargs):
        """ """
        direction = kwargs.get("direction", "apical_trunk")
        if direction == "apical_trunk":
            sec_name = kwargs.get("seclist_name", "apical")

            if sec_name != "apical":
                raise ValueError("With direction 'apical_trunk', sec_name must be apical")

            location = NrnTrunkSomaDistanceCompLocation(
                name="dend",
                soma_distance=kwargs["somadistance"],
                sec_index=kwargs.get("sec_index", None),
                seclist_name="apical",
            )
        elif direction == "random":
            location = NrnSomaDistanceCompLocation(
                name="dend",
                soma_distance=kwargs["somadistance"],
                seclist_name=kwargs.get("seclist_name", "apical"),
            )
        else:
            raise ValueError(f"direction keyword {direction} not understood")
        super().__init__(location=location, **kwargs)

    def instantiate(self, sim=None, icell=None):
        """Force to have holding current at 0."""
        self.holding_current = 0
        super().instantiate(sim=sim, icell=icell)


class Synaptic(DendriticStep):
    """Ecode to model a synapse with EPSP-like shape.

    A synthetic EPSP shape is defined by the difference of two exponentials, one with a
    rise time (syn_rise) constant, the other with a decay (syn_decay) time constants.
    It is normalized such that the maximum value is parametrized by syn_amp.
    """

    name = "Synaptic"

    def __init__(self, location, **kwargs):
        """Constructor

        Args:
            step_amplitude (float): amplitude (nA)
            step_delay (float): delay (ms)
            step_duration (float): duration (ms)
            location (Location): stimulus Location
            syn_delay (float): start time of synaptic input
            syn_amp (flaot): maximal amplitude of the synaptic input
            syn_rise (float): rise time constant
            syn_decay (float): decay time constant
        """
        super().__init__(location=None, **kwargs)

        self.syn_delay = kwargs.get("syn_delay", 0.0)
        self.syn_amp = kwargs.get("syn_amp", 0.0)
        self.syn_rise = kwargs.get("syn_rise", 0.5)
        self.syn_decay = kwargs.get("syn_decay", 5.0)

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""
        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.total_duration

        self.current_vec = sim.neuron.h.Vector()
        self.time_vec = sim.neuron.h.Vector()
        self.time_vec.append(0.0)
        self.current_vec.append(0)

        self.time_vec.append(self.delay + self.syn_delay)
        self.current_vec.append(0)

        t = np.linspace(0, self.total_duration - self.delay - self.syn_delay, 2000)
        s = np.exp(-t / self.syn_decay) - np.exp(-t / self.syn_rise)
        s = self.syn_amp * s / max(s)

        for _t, _s in zip(t, s):
            self.time_vec.append(self.delay + self.syn_delay + _t)
            self.current_vec.append(_s)

        self.iclamp.delay = 0
        self.current_vec.play(
            self.iclamp._ref_amp,  # pylint:disable=W0212
            self.time_vec,
            1,
            sec=icomp.sec,
        )


class BAC(IDrest):
    """BAC ecode.

    BAC is a combination of a bAP and a synaptic input to generate Ca dendritic spikes.
    """

    def __init__(self, location, **kwargs):
        """Constructor, combination of IDrest and Synaptic ecodes."""
        self.bap = IDrest(location, **kwargs)
        self.epsp = Synaptic(None, **kwargs)

        super().__init__(location=None, **kwargs)

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""
        self.bap.holding_current = self.holding_current
        self.bap.threshold_current = self.threshold_current
        self.bap.instantiate(sim=sim, icell=icell)
        self.epsp.instantiate(sim=sim, icell=icell)



"""E-model access_point module"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


def get_access_point(access_point, emodel, **kwargs):
    """Returns a DataAccessPoint object.

    Args:
        access_point (str): name of the access_point to use, can be 'nexus' or 'local'.
        emodel (str): name of the emodel.
        kwargs: extra arguments to pass to access_point constructors, see below.

    Optional:
        - etype (str): name of the electric type.
        - ttype (str): name of the transcriptomic type.
        - mtype (str): name of the morphology type.
        - species (str): name of the species.
        - brain_region (str): name of the brain location.
        - iteration (str): tag associated to the current run.
        - synapse_class (str): synapse class (neurotransmitter).

    For local:
        - emodel_dir (str): path of the directory containing the parameters,
                            features and parameters config files.
        - recipe_path (str, optional): path to the file containing the recipes.
        - final_path (str, optional): path to the final.json,
                                      if different from the one in emodel_dir
        - legacy_dir_structure (bool, optional): uses legacy folder structure
        - with_seed (bool): allows for emodel_seed type of emodel names in final.json
                            (not in recipes)

    For nexus:
        - project (str): name of the Nexus project.
        - organisation (str): name of the Nexus organization to which the project belong.
        - endpoint (str): Nexus endpoint.
        - forge_path (str): path to a .yml used as configuration by nexus-forge.
        - forge_ontology_path (str): path to the .yml used to access the ontology in Nexus Forge.
                                     If not provided, forge_path will be used.
        - access_token (str, optional): Nexus connection token.

    Returns:
        DataAccessPoint
    """

    etype = kwargs.get("etype", None)
    etype = etype.replace("__", " ") if etype else None

    ttype = kwargs.get("ttype", None)
    ttype = ttype.replace("__", " ") if ttype else None

    mtype = kwargs.get("mtype", None)
    mtype = mtype.replace("__", " ") if mtype else None

    brain_region = kwargs.get("brain_region", None)
    brain_region = brain_region.replace("__", " ") if brain_region else None

    if access_point == "nexus":
        try:
            from bluepyemodelnexus.nexus import NexusAccessPoint
        except ImportError as exc:
            raise ImportError(
                "The internal bluepyemodelnexus package is required to use the Nexus access point."
            ) from exc

        return NexusAccessPoint(
            emodel=emodel,
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=kwargs.get("species", None),
            brain_region=brain_region,
            iteration_tag=kwargs.get("iteration_tag", None),
            synapse_class=kwargs.get("synapse_class", None),
            project=kwargs.get("project", None),
            organisation=kwargs.get("organisation", "bbp"),
            endpoint=kwargs.get("endpoint", "https://bbp.epfl.ch/nexus/v1"),
            forge_path=kwargs.get("forge_path", None),
            forge_ontology_path=kwargs.get("forge_ontology_path", None),
            access_token=kwargs.get("access_token", None),
        )

    if access_point == "local":
        from bluepyemodel.access_point.local import LocalAccessPoint

        return LocalAccessPoint(
            emodel=emodel,
            emodel_dir=kwargs.get("emodel_dir", None),
            etype=etype,
            ttype=ttype,
            mtype=mtype,
            species=kwargs.get("species", None),
            brain_region=brain_region,
            iteration_tag=kwargs.get("iteration_tag", None),
            synapse_class=kwargs.get("synapse_class", None),
            recipes_path=kwargs.get("recipes_path", None),
            final_path=kwargs.get("final_path", None),
            legacy_dir_structure=kwargs.get("legacy_dir_structure", False),
            with_seeds=kwargs.get("with_seeds", False),
        )

    raise ValueError(f"Unknown access point: {access_point}. Should be 'nexus' or 'local'.")



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../../."))

from importlib.metadata import version


# -- Project information -----------------------------------------------------

project = "BluePyEModel"
author = "Blue Brain Project/EPFL"

# The short X.Y version
version = version("bluepyemodel")

# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.graphviz",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx-bluebrain-theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# html_theme_options = {
#    "metadata_distribution": "BluepyEModel",
# }

html_title = "BluePyEModel"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# autosummary settings
autosummary_generate = True

# autodoc settings
autodoc_typehints = "signature"
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
}
autoclass_content = "both"

add_module_names = False

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "luigi": ("https://luigi.readthedocs.io/en/stable", None),
}

autodoc_mock_imports = ["bluepyemodel.tasks.luigi_tools"]



"""TraceFile"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

logger = logging.getLogger(__name__)


def list_ecodes_per_traces(traces, threshold_count=0):
    """Utility method. Return the list of ecodes available for each trace. Only
    the ecodes present more than threshold_count times are considered."""

    ecodes_per_traces = {}
    all_ = []
    for t in traces:
        if t.ecodes:
            ecodes = [str(e) for e in t.ecodes]
            ecodes_per_traces[t.cell_name] = ecodes
            all_ += ecodes
        else:
            ecodes_per_traces[t.cell_name] = []

    ecodes_per_traces["all"] = sorted(list(set(all_)))

    count_ecodes = {e: all_.count(e) for e in ecodes_per_traces["all"]}

    for t in ecodes_per_traces.copy():
        ecodes_per_traces[t] = list(
            set(e for e in ecodes_per_traces[t] if e and count_ecodes[e] > threshold_count)
        )

    return ecodes_per_traces


class TraceFile:
    """Contains the metadata of a trace file"""

    def __init__(
        self,
        cell_name,
        filename=None,
        filepath=None,
        resource_id=None,
        ecodes=None,
        other_metadata=None,
        species=None,
        brain_region=None,
        etype=None,
        id=None,
    ):
        """Constructor.

        Args:
            cell_name (str): cell name
            filename (str): file name
            filepath (str): file path
            resource_id (str): deprecated. kept for legacy reasons
            ecodes (dict): the ecodes, e.g.

                .. code-block::

                    {
                        "IDrest": {
                            "t_unit": "s",
                            "ljp": 14.0,
                            "protocol_name": "IDrest"
                    }
            other_metadata (dict): voltage and current files, e.g.

                .. code-block::

                    {
                        "v_file": "path_to_voltage_file",
                        "i_file": "path_to_current_file"
                    }

            species (dict): species, e.g.

                .. code-block::

                    {
                        "id": "http://purl.obolibrary.org/obo/NCBITaxon_10090",
                        "label": "Mus musculus",
                    }
            brain_region (dict): brain region, e.g.

                .. code-block::

                    {
                        "@id": "http://api.brain-map.org/api/v2/data/Structure/407",
                        "label": "Field CA1, pyramidal layer"
                    }
            etype (str): etype
            id (str): the nexus id of the trace resource
        """
        self.cell_name = cell_name
        self.filename = filename if filename else cell_name
        self.filepath = filepath
        self.resource_id = resource_id

        self.ecodes = ecodes

        self.other_metadata = other_metadata if other_metadata is not None else {}

        self.species = species
        self.brain_region = brain_region
        self.etype = etype

        self.id = id

    def as_dict(self):
        return vars(self)

    def __eq__(self, other):
        if self.cell_name == other.cell_name:
            if self.filename and other.filename:
                if self.filename == other.filename:
                    return True
                return False

            return True

        return False



"""BPEM_stimulus class"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

import numpy
from bluepyopt.ephys.stimuli import NrnSquarePulse

logger = logging.getLogger(__name__)


class BPOSquarePulse(NrnSquarePulse):
    """Abstract current stimulus based on BluePyOpt square stimulus.

    Can be used to reproduce results using BluePyOpt's NrnSquarePulse.

    .. code-block:: none

              holdi               holdi+amp                holdi
                :                     :                      :
                :                     :                      :
                :           ______________________           :
                :          |                      |          :
                :          |                      |          :
                :          |                      |          :
                :          |                      |          :
        |__________________|                      |______________________
        ^                  ^                      ^                      ^
        :                  :                      :                      :
        :                  :                      :                      :
        t=0                delay                  delay+duration         totduration
    """

    name = ""

    def __init__(self, location, **kwargs):
        """Constructor
        Args:
            location(Location): location of stimulus
        """

        super().__init__(
            step_amplitude=kwargs.get("amp", None),
            step_delay=kwargs.get("delay", 250.0),
            step_duration=kwargs.get("duration", 1350.0),
            total_duration=kwargs.get("totduration", 1850.0),
            location=location,
        )

        self.holding_current = kwargs.get("holding_current", None)
        self.holding_iclamp = None

    @property
    def stim_start(self):
        return self.step_delay

    @property
    def stim_end(self):
        return self.step_delay + self.step_duration

    @property
    def amplitude(self):
        return self.step_amplitude

    def instantiate(self, sim=None, icell=None):
        """Run stimulus"""

        icomp = self.location.instantiate(sim=sim, icell=icell)

        self.iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
        self.iclamp.dur = self.step_duration

        self.iclamp.delay = self.step_delay
        self.iclamp.amp = self.step_amplitude

        if self.holding_current is not None:
            self.holding_iclamp = sim.neuron.h.IClamp(icomp.x, sec=icomp.sec)
            self.holding_iclamp.dur = self.total_duration
            self.holding_iclamp.delay = 0
            self.holding_iclamp.amp = self.holding_current

    def destroy(self, sim=None):  # pylint:disable=W0613
        """Destroy stimulus"""
        self.iclamp = None
        self.holding_iclamp = None

    def generate(self, dt=0.1):
        """Return current time series"""
        holding_current = self.holding_current if self.holding_current is not None else 0

        t = numpy.arange(0.0, self.total_duration, dt)
        current = numpy.full(t.shape, holding_current, dtype="float64")

        ton_idx = int(self.stim_start / dt)
        toff_idx = int(self.stim_end / dt)

        current[ton_idx:toff_idx] += self.amplitude

        return t, current

    def __str__(self):
        """String representation"""
        return f"{self.name} current played at {self.location}"



"""Model selector selects NEURON mechanisms for cell model configuration"""

"""
Copyright 2023, EPFL/Blue Brain Project

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import copy

from .mechanism import Mechanism


class ModelSelector:
    """Selects NEURON mechanisms for cell model configuration"""

    def __init__(self, ic_map):
        """
        Args:
            ic_map (dict): content of the icmapping.json file
        """

        self._ic_map = ic_map
        self._mechanisms = self._load_mechanisms()
        self.status = "stable"
        self.mode = "mixed"

    def _get_ic_map_entry(self, key):
        """Get an entry from the icmap

        Args:
            key (str): name of the field

        Returns:
            entry: content of the field
        """

        entry = self._ic_map[key]
        entry = {k: v for k, v in entry.items() if not k[0] == "_"}
        return entry

    def _load_mechanisms(self):
        """Load mechanism field from the icmap

        Returns:
            mechanisms (dict [Mechanism]):
        """
        ic_map_mechs = self._get_ic_map_entry("mechanisms")
        mechanisms = {k: Mechanism(**v) for k, v in ic_map_mechs.items()}
        return mechanisms

    def _set_mechanism_parameters(self):
        """Set mechanism parameters from icmap settings."""

        mechs = self._mechanisms
        params = self._get_ic_map_entry("mechanism_parameters")
        for name, info in params.items():
            mech = mechs[name]
            mech.set_from_icmap(info)

    def _get_mech_from_channel(self, channel_name):
        """Get mechanism corresponding to channel

        Args:
            channel_name (str): name of the ion channel

        Returns:
            mech (Mechanism): corresponding mechanism
        """

        chan_to_model = self._get_ic_map_entry("channels")
        models = chan_to_model.pop(channel_name, [])
        mechs = [self._get_mech_from_model(model) for model in models]

        for mech in mechs:
            if self._check_mech(mech):
                return mech
        return None

    def _check_mech(self, mech):
        if not mech:
            return False

        if hasattr(mech, "type"):
            return self.mode in (mech.type, "mixed")
        return True

    def _get_mech_from_model(self, model_name):
        """Get mechanism by model name

        Args:
            model_name (str): model name

        Returns:
            (Mechanism): corresponding mechanism
        """

        if model_name in self._mechanisms:
            mech = self._mechanisms[model_name]
            if self._check_mech(mech):
                return mech
        return None

    def get(self, name):
        """Get mechanism from channel or model name
        Args:
            name (str): channel or model name
        Returns:
            (Mechanism): corresponding mechanism
        """

        mech = self._get_mech_from_channel(name)
        if not mech:
            mech = self._get_mech_from_model(name)
        return mech

    def select(self, name):
        """Select and return mechanism by channel or model name
        Args:
            name (str): channel or model name
        Returns:
            (Mechanism): corresponding mechanism
        """

        mech = self.get(name)
        if not mech:
            return None
        mech.select(check_status=self.status)
        if mech.is_selected():
            for req_mod in mech.requires:
                req_mech = self.select(req_mod)  # NB: recursion!
                if req_mech.is_selected():
                    req_mech.set_distribution(mech.distribution)
        return mech

    def get_mechanisms(self, selected_only=True):
        """Get a copy of all available mechanisms from the icmapping file.
        Args:
            selected_only (bool): flag to get only selected channels
        Returns:
            mechs (dict [Mechanism]): mechanisms with all associated info
                fields
        """

        self.select("pas")
        self._set_mechanism_parameters()
        if selected_only:
            mechs = {name: mech for name, mech in self._mechanisms.items() if mech.is_selected()}
        else:
            mechs = self._mechanisms
        return copy.deepcopy(mechs)

    def __str__(self):
        out_str = []
        out_str += ["\n>>> Mechanisms <<<"]
        n = 1
        mechs = self.get_mechanisms()
        for v in mechs.values():
            items_str = f"-{n}- {str(v)}"
            n += 1
            out_str.append(items_str)
        return "\n".join(out_str)
